      SUBROUTINE ALGO2(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XQ,
     &                 IPAR, RPAR, RWORK, LRWORK, IERR)
C     .. Scalar Arguments..
      INTEGER          M, N, NSTEP, LRWORK, IERR
	DOUBLE PRECISION T0, TN
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION XQ(*), RPAR(*), RWORK(*)     
C     .. Subroutine Arguments ..
      EXTERNAL         EDIF, ADIF, FDIF
C
C  Purpose:
C 
C   This subroutine computes the solution of the minimization problem
C                            .
C      ||x||_2=min!  s.t. ||Ex-Ax-f||_2=min!,
C
C   where
C       .
C      Ex = Ax + f, x(t_0)=0 
C
C   is a strangeness-free linear differential algebraic equation in
C   orthogonal standard form. This solution is computed by solving
C   the minimization problem
C                            
C      ||x_h||_2=min!  s.t. ||D_h*x_h-f_h||_2=min!
C
C   where the linear system
C
C      D_h*x_h = f_h
C
C   represents the discretization of the above differential algebraic
C   system with the implicit Euler method and with a homogegeous 
C   initial value.
C
C   
C       ARGUMENTS IN
C
C         M - INTEGER
C             The number of equations in the DAE system.
C             M .GE. 1.
C
C         N - INTEGER.
C             The number of variables in the DAE system.
C             N .GE. 1.
C
C     NSTEP - INTEGER
C             The number of discretization steps.
C             NSTEP .GE. 1.
C
C      IPAR - INTEGER array of DIMENSION (*)
C             Parameter array. If IPAR(1)=0, then the subroutines
C             EDEF and ADEF define the coefficient matrices E(t) and
C             A(t)  of a DAE system in orthogonal standard form. 
C             If IPAR(1)=1, then EDEF and ADEF define the matrices
C             E(t)*Q(t) and A(t)*Q(t)-E(t)*(d/dt)Q(t) where the smooth
C             orthogonal transformation Q and its derivative are defined
C             by the subroutines QDEF and QDOT.
C
C      IOPT - INTEGER array of DIMENSION 1
C             Option array. Up to now there is just one option to choose.
C             If IOPT(1).EQ.0 then the code assumes that the pair (E,A) 
C             has pointwise full row rank.
C             If IOPT(1).EQ.1 then the code computes the rank NEQ of 
C             the pairs (E(t_i),A(t_i)) and transformation U_i at every
C             timestep t_i=T0+i*H for i=1,...,NSTEP such that the first
C             NEQ rows of the pair (U*E(t_i),U*A(t_i)) are linearly
C             independent. If NEQ changes, the computation is cancelled
C             and an error message is given.
C
C
C       ARGUMENTS OUT
C
C        XQ - DOUBLE PRECISION array of DIMENSION (NSTEP*N)
C             For i=1:NSTEP XQ((i-1)+1:i*N) contains the 
C             approximation x_i to the solution of the discrete
C             minimization problem at the timestep t_i=t_0+i*h 
C             computed by this subroutine.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - NTEGER.
C              The length of RWORK.
C              LRWORK .GE. NEQ*(N + 4*NEQ) + N + M*(1 + 2*N + M)
C                          + MAX(3*M,2*N)+ NSTEP*NEQ*(3 + 2*N + 2*NEQ)
C              where NEQ.LE.M denotes the rank of the matrix pair 
C              (E(t_i),A(t_i)) at t_i=T0+i*H for i=1,...,NSTEP. 
C
C
C     The following subroutines have to be provided by the user:
C
C    SETSIZE - User supplied SUBROUTINE
C              This subroutine defines the size of the DAE system and
C              the characteristic values \hat d, \hat a, \hat u and
C              \hat v. It has the form
C                SUBROUTINE SETSIZE(M, N, ID, IA, IU, IV).
C              As output, this subroutine produces the number M of 
C              equations and the number N of variables of the 
C              strangeness-free DAE system. Furthermore it defines 
C              the characteristic values ID=\hat d, IA=\hat a, 
C              IU=\hat u and IV=\hat v of the DAE system.
C
C     SETINT - User supplied SUBROUTINE
C              This subroutine defines the left and right bounds of
C              the integration interval [t_0, T]. It has the form
C                SUBROUTINE SETINT(T0, TN).
C              As output, this subroutine defines the left and the 
C              right interval bounds T0=t_0 and TN=T.
C       
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C              If IPAR(1)=0, then the matrix E has to be defined 
C              according to the orthogonal standard form, i.e. the 
C              leading ID by ID part of E(t) has to be nonsingular,
C              the rest of the matrix has to be equal to zero.
C              If IPAR(1)=1, then the array E contains the matrix
C              E(t)*Q(t) where the orthogonal matrix Q(t) is defined by
C              the subroutine QDEF.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C              If IPAR(1)=0, then the matrix A has to be defined 
C              according to the orthogonal standard form, i.e. 
C              only the blocks A(1:ID,1:N) and A(ID+1:ID+IA,1:ID+IA)
C              contain nonzero entries and the block 
C              A(ID+1,ID+IA,ID+1:ID+IA) is nonsingular.
C              If IPAR(1)=1, then the array A contains the matrix
C              A(t)*Q(t)-E(t)*(d/dt)Q(t) where the orthogonal matrix 
C              Q(t) and its derivative (d/dt)Q(t) are defined by
C              the subroutines QDEF and QDOT.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C     ERROR INDICATOR
C
C      IERR - INTEGER.
C             If IERR.LT.0, then the call to the LAPACK subroutine DGELS
C             failed.
C
C
C
C     .. Local Scalars ..
      INTEGER          I, J, I2, LDA, LDE, LDDH, OA, OE, OF, ODH, OXH,
     &                 OWORK, LWORK
      DOUBLE PRECISION H
C     .. Executable Statements ..
C
C   Setup work-array
C
      LDA   = M
      LDE   = M
	LDDH  = M*NSTEP
      OE    = 1
      OA    = OE + LDE*N
	OF    = OA + LDA*N
	ODH   = OF + 2*M
	OXH   = ODH + LDDH*NSTEP*N
      OWORK = OXH + N*NSTEP
	LWORK = LRWORK - OWORK
	IF (LWORK .LE. NSTEP*(M+N)) THEN
         WRITE (6,100) LRWORK, OWORK+NSTEP*(M+N)
	   RETURN
      END IF 
  100 FORMAT (1X,'DGMIN: LRWK=',i8,
     $        ' is too small! Set LRWK at least to',i8)
C
C   Setup the discretization matrix and solve the system
C
      CALL SETDH(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, IPAR, 
     &           RPAR, RWORK(OE), LDE, RWORK(OA), LDA, RWORK(OF),
     &           RWORK(ODH), LDDH, RWORK(OXH), RWORK(OWORK), LRWORK)
      CALL DGELS('N', M*NSTEP, N*NSTEP, 1, RWORK(ODH), LDDH, 
     &           RWORK(OXH), N*NSTEP, RWORK(OWORK), LWORK, IERR)
      RETURN
      END
