      SUBROUTINE DGLOBAL(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XQ,
     &                   CVAL, IPAR, RPAR, RWORK, LRWORK, IOPT, IERR)
C     .. Scalar Arguments..
      INTEGER          M, N, NSTEP, LRWORK, IERR
	DOUBLE PRECISION T0, TN
C     .. Array Arguments ..
      INTEGER          CVAL(*), IPAR(*), IOPT(*)
      DOUBLE PRECISION XQ(*), RPAR(*), RWORK(*)     
C     .. Subroutine Arguments ..
      EXTERNAL         EDIF, ADIF, FDIF
C
C  Purpose:
C 
C   This subroutine computes the solution of the minimization problem
C                            .
C      ||x||_2=min!  s.t. ||Ex-Ax-f||_2=min!,
C
C   where
C       .
C      Ex = Ax + f, x(t_0)=0 
C
C   is a strangeness-free linear differential algebraic equation in
C   orthogonal standard form. This solution is computed by solving
C   the minimization problem
C                            
C      ||x_h||_2=min!  s.t. ||D_h*x_h-f_h||_2=min!
C
C   where the linear system
C
C      D_h*x_h = f_h
C
C   represents the discretization of the above differential algebraic
C   system with the implicit Euler method and with a homogegeous 
C   initial value. This system is solved with respect to the 
C   block bidiagonal structure of the matrix D_h.
C
C       USER-SUPPLIED SUBROUTINES
C
C      EDIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             matrix E(t) and its derivatives. It has the form
C
C             SUBROUTINE EDIF(M,N,T,IDIF,E,LDE,IPAR,RPAR,IERR).
C
C             The subroutine takes as input the number of equations M,
C             the number of variables N,
C             the time T and the integer parameter IDIF. EDIF must not
C             alter these input parameters or the leading dimension LDE
C             of the array E. As output, the subroutine produces the
C             IDIF-th derivative of E(t) at time T in the leading N by N
C             part of the array E. The integer flag IERR is always zero
C             on input and EDIF should alter IERR only if IDIF is larger
C             than the highest derivative of E(t) the subroutine
C             provides (set IERR=-2) or if another problem occurs (set
C             IERR=-1). IPAR and RPAR are integer and real arrays which
C             can be used for the communication between the user's
C             calling program and the subroutines EDIF, ADIF and FDIF.
C
C             In the calling program, EDIF must be declared as external.
C
C      ADIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             matrix A(t) and its derivatives. It is of the form
C
C              SUBROUTINE ADIF(M,N,T,IDIF,A,LDA,IPAR,RPAR,IERR)
C
C             and the input and output parameters are similar to these
C             of EDIF.
C
C      FDIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             vector f(t) and its derivatives. It is of the form
C
C             SUBROUTINE FDIF(M,N,T,IDIF,F,IPAR,RPAR,IERR)
C
C             and the input and output parameters are similar to these
C             of EDIF. Except that the first N elements of the
C             1-dimensional array F contain the IDIF-th derivative of
C             f(t) at time T. Note further, since F is a 1-dimensional
C             array no leading dimension is needed.
C
C       ARGUMENTS IN
C
C         M - INTEGER
C             The number of equations in the DAE system.
C             M .GE. 1.
C
C         N - INTEGER.
C             The number of variables in the DAE system.
C             N .GE. 1.
C
C        T0 - DOUBLE PRECISION
C             The left bound of the integration interval
C
C        TN - DOUBLE PRECISION
C             The right bound of the integration interval
C
C     NSTEP - INTEGER
C             The number of discretization steps.
C             NSTEP .GE. IORD+1.
C
C      IPAR - INTEGER array of DIMENSION (*).
C             This integer array can be used for communication between
C             the calling program and the EDIF, ADIF and FDIF
C             subroutines.
C
C      RPAR - DOUBLE PRECISION array of DIMENSION (*).
C             This real array can be used for communication between the
C             calling program and the EDIF, ADIF and FDIF subroutines.
C
C             IPAR and RPAR are not altered by DGLOBAL or its
C             subprograms. If these arrays are used, they must be
C             dimensioned in the calling program and in EDIF, ADIF and
C             FDIF as arrays of appropriate length. Otherwise, ignore
C             these arrays by treating them as dummy arrays of length
C             one.
C
C
C       ARGUMENTS OUT
C
C        XQ - DOUBLE PRECISION array of DIMENSION (NSTEP*N)
C             For i=1:NSTEP X((i-1)+1:i*N) contains the 
C             approximation x_i to the solution of the DAE system
C             at the timestep t_i=t_0+i*h computed by this subroutine.
C
C      CVAL - INTEGER array of DIMENSION (5)
C             If INFO(1)=0 this array contains the characteristic
C             values of the DAE
C             CVAL(1) contains the strangeness index MU.
C             CVAL(2) contains the number DMU of differential
C                     components.
C             CVAL(3) contains the number AMU of algebraic components.
C             CVAL(4) contains the number UMU of undetermined
C                     components.
C             CVAL(5) contains the number VMU of solvability
C                     conditions.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - INTEGER.
C              The length of RWORK.
C              LRWORK .GE. MIN(M,N) + MAX(M,N).
C
C     MODE PARAMETERS
C
C      IOPT - INTEGER array of DIMENSION 1.
C             IOPT(1) If the equivalent strangeness free system should
C                     be computed before the discretization, set 
C                     IOPT(1) = 0, otherwise it is assumed that the 
C                     system is strangeess free and of full row rank.
C
C     ERROR INDICATOR
C
C      IERR - INTEGER.
C             If IERR.LT.0, then the call to the LAPACK subroutine DGELS
C             or the computation of the reduced system failed.
C
C
C     .. Local Scalars ..
      INTEGER          I, J, I2, LDA, LDE, LDT, LDTU, LDQ1, LDQ2,
     &                 LDQ3, OA, OE, OF, OT, OTU, OU, OXQ, OEL, OFQ,
     &                 OQ1, OQ2, OQ3, OTAU1, OTAU2, OTAU3, OWORK, 
     &                 OSIGMA, LWORK, NEQ 
      DOUBLE PRECISION H
C     .. Executable Statements ..
C
C   Compute stepsize and setup work-array
C
	MAXIND = 1
	MQ = MAXIND*M
	NQ = MAXIND*N
      H=(TN-T0)/NSTEP
C
C   Setup work-array
C
      LDA   = M
      LDE   = M
      OE    = 1
      OA    = OE + LDE*N
	OF    = OA + LDA*N
      OWORK = OF + 2*M
	LWORK = LRWORK - OWORK
	IF (LWORK .LE. MIN(3*M, 2*N)) THEN
         WRITE (6,100) LRWORK, OWORK+MIN(3*M, 2*N)
	   RETURN
      END IF 
  100 FORMAT (1X,'DGMIN: LRWK=',i8,
     $        ' is too small! Set LRWK at least to',i8)      
C
C   Compute rank of [E(TN),A(TN)] if necessary
C 
	IF (IOPT(1) .EQ. 0) THEN
	   CALL REDUCE(EDIF, ADIF, FDIF, M, N, TN, MAXIND, CVAL,
     $               RWORK(OE), LDE, RWORK(OA), LDA, RWORK(OF), IPAR,
     $               RPAR, RWORK(OWORK), LWORK, IERR)
	   NEQ = CVAL(2) + CVAL(3)
      ELSE
	   CALL EDIF(M, N, TN, 0, RWORK(OE), M, IPAR, RPAR, IERR)
	   CALL ADIF(M, N, TN, 0, RWORK(OA), M, IPAR, RPAR, IERR)
	   CALL FDIF(M, TN, 0, RWORK(OF), IPAR, RPAR, IERR)
	   NEQ = M
	END IF
C
C   Update work-array
C
      LDT   = NEQ
      LDEL  = NEQ
      LDTU  = NEQ
      LDQ1  = NSTEP*NEQ
      LDQ2  = NSTEP*NEQ
      LDQ3  = NSTEP*NEQ
	OT    = OWORK
      OEL   = OT + LDT*N
      OTU   = OEL + LDEL*2*NEQ
      OQ1   = OTU + LDTU*2*NEQ
      OQ2   = OQ1 + LDQ1*N
      OQ3   = OQ2 + LDQ2*N
      OTAU1 = OQ3 + 2*LDQ3*NEQ
      OTAU2 = OTAU1 + NSTEP*NEQ
      OTAU3 = OTAU2 + NSTEP*NEQ
      OWORK = OTAU3 + NSTEP*NEQ + NEQ
      LWORK = LRWORK - OWORK
C
C   T = E - hA, E <-- -E 
C    
      DO 200 I=1,NEQ
         DO 200 J=1,N
	    IPOSE = OE + (I-1) + (J-1)*LDE
	    IPOSA = OA + (I-1) + (J-1)*LDA
	    IPOST = OT + (I-1) + (J-1)*LDT
	    RWORK(IPOST) = RWORK(IPOSE) - H*RWORK(IPOSA)
	    RWORK(IPOSE) = -RWORK(IPOSE)
 200  CONTINUE
C
C   Compute the least square solution
C
      CALL DECOMP(M, N, NSTEP, T0, H, XQ, CVAL, RWORK(OF), RWORK(OE),
     &            LDE, RWORK(OA), LDA, RWORK(OT), LDT, RWORK(OEL), 
     &            LDEL, RWORK(OTU), LDTU, RWORK(OQ1), LDQ1, RWORK(OQ2),
     &            LDQ2, RWORK(OQ3), LDQ3, RWORK(OTAU1), RWORK(OTAU2),
     &            RWORK(OTAU3), RWORK(OWORK), LWORK, EDIF, ADIF, FDIF,
     &            IPAR, RPAR, MAXIND, IOPT) 
      RETURN
      END
