      SUBROUTINE DLOCAL(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XQ, CVAL,
     $                  IPAR, RPAR, RWORK, LRWORK, IOPT, IERR)
C     .. Scalar Arguments..
      INTEGER          M, N, NSTEP, LRWORK, IORD, MAXIND
	DOUBLE PRECISION T0, TN
C     .. Array Arguments ..
      INTEGER          CVAL(*), IPAR(*), IOPT(*)
      DOUBLE PRECISION XQ(*), RPAR(*), RWORK(*)
C     .. Subroutine Arguments
      EXTERNAL         EDIF, ADIF, FDIF
C
C  Purpose:
C   
C  Computes a solution of an over- or underdetermined linear 
C  strangeness-free differential algebraic equation using the
C  implicit Euler method with a constant step size. In every step of
C  the  discretization the least-square solution of the occuring 
C  linear system is chosen as the next iterate.
C
C       USER-SUPPLIED SUBROUTINES
C
C      EDIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             matrix E(t) and its derivatives. It has the form
C
C             SUBROUTINE EDIF(M,N,T,IDIF,E,LDE,IPAR,RPAR,IERR).
C
C             The subroutine takes as input the number of equations M,
C             the number of variables N,
C             the time T and the integer parameter IDIF. EDIF must not
C             alter these input parameters or the leading dimension LDE
C             of the array E. As output, the subroutine produces the
C             IDIF-th derivative of E(t) at time T in the leading N by N
C             part of the array E. The integer flag IERR is always zero
C             on input and EDIF should alter IERR only if IDIF is larger
C             than the highest derivative of E(t) the subroutine
C             provides (set IERR=-2) or if another problem occurs (set
C             IERR=-1). IPAR and RPAR are integer and real arrays which
C             can be used for the communication between the user's
C             calling program and the subroutines EDIF, ADIF and FDIF.
C
C             In the calling program, EDIF must be declared as external.
C
C      ADIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             matrix A(t) and its derivatives. It is of the form
C
C              SUBROUTINE ADIF(M,N,T,IDIF,A,LDA,IPAR,RPAR,IERR)
C
C             and the input and output parameters are similar to these
C             of EDIF.
C
C      FDIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             vector f(t) and its derivatives. It is of the form
C
C             SUBROUTINE FDIF(M,N,T,IDIF,F,IPAR,RPAR,IERR)
C
C             and the input and output parameters are similar to these
C             of EDIF. Except that the first N elements of the
C             1-dimensional array F contain the IDIF-th derivative of
C             f(t) at time T. Note further, since F is a 1-dimensional
C             array no leading dimension is needed.
C
C       ARGUMENTS IN
C
C         M - INTEGER
C             The number of equations in the DAE system.
C             M .GE. 1.
C
C         N - INTEGER.
C             The number of variables in the DAE system.
C             N .GE. 1.
C
C        T0 - DOUBLE PRECISION
C             The left bound of the integration interval
C
C        TN - DOUBLE PRECISION
C             The right bound of the integration interval
C
C     NSTEP - INTEGER
C             The number of discretization steps.
C             NSTEP .GE. IORD+1.
C
C      IPAR - INTEGER array of DIMENSION (*).
C             This integer array can be used for communication between
C             the calling program and the EDIF, ADIF and FDIF
C             subroutines.
C
C      RPAR - DOUBLE PRECISION array of DIMENSION (*).
C             This real array can be used for communication between the
C             calling program and the EDIF, ADIF and FDIF subroutines.
C
C             IPAR and RPAR are not altered by DGLOBAL or its
C             subprograms. If these arrays are used, they must be
C             dimensioned in the calling program and in EDIF, ADIF and
C             FDIF as arrays of appropriate length. Otherwise, ignore
C             these arrays by treating them as dummy arrays of length
C             one.
C
C
C       ARGUMENTS OUT
C
C        XQ - DOUBLE PRECISION array of DIMENSION (NSTEP*N)
C             For i=1:NSTEP X((i-1)+1:i*N) contains the 
C             approximation x_i to the solution of the DAE system
C             at the timestep t_i=t_0+i*h computed by this subroutine.
C
C      CVAL - INTEGER array of DIMENSION (5)
C             If INFO(1)=0 this array contains the characteristic
C             values of the DAE
C             CVAL(1) contains the strangeness index MU.
C             CVAL(2) contains the number DMU of differential
C                     components.
C             CVAL(3) contains the number AMU of algebraic components.
C             CVAL(4) contains the number UMU of undetermined
C                     components.
C             CVAL(5) contains the number VMU of solvability
C                     conditions.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - INTEGER.
C              The length of RWORK.
C              LRWORK .GE. MIN(M,N) + MAX(M,N).
C
C     MODE PARAMETERS
C
C      IOPT - INTEGER array of DIMENSION 2.
C             IOPT(1) If the equivalent strangeness free system should
C                     be computed before the discretization, set 
C                     IOPT(1) = 0, otherwise it is assumed that the 
C                     system is strangeess free and of full row rank.
C             IOPT(2) Set IOPT(2) = 0 if the implicit Euler method 
C                     should be used for the discretization.
C                     Set IOPT(2) = k if the code should use a k-step
C                     BDF method.
C
C     ERROR INDICATOR
C
C      IERR - INTEGER.
C             If IERR.LT.0, then the call to the LAPACK subroutine DGELS
C             or the computation of the reduced system failed.
C
C
C     .. Local Scalars ..
      INTEGER          I, ISTEP, IO, NEQ
      DOUBLE PRECISION H, T
C     .. Local Arrays ..
      DOUBLE PRECISION COEFF(0:6,1:6), F(MAX(M,N)), E(M,N), A(M,N)
C     .. External Subroutines ..
      EXTERNAL         GETCOF, DSCAL, DGEMV, DCOPY, DGELS, DLASET
C     .. Executable Statements ..
      H = (TN-T0)/NSTEP
	IF (IOPT(2) .EQ. 0) THEN
	   IORD = 1
	ELSE
	   IORD = IOPT(2)
	END IF
	CALL GETCOF(COEFF)
      T = T0 + IORD*H
	IRED = IOPT(1)
C
C   Compute the coefficients of the stangeness free system at t=t_0+h.
	IF (IRED .EQ.0) THEN
	   CALL REDUCE(EDIF, ADIF, FDIF, M, N, T, MAXIND, CVAL, E, M, A,
     $               M, F, IPAR, RPAR, RWORK, LRWORK, IERR)
	   NEQ = CVAL(2) + CVAL(3)
	ELSE
	   CALL EDIF(M, N, T, 0, E, M, IPAR, RPAR, IERR)
	   CALL ADIF(M, N, T, 0, A, M, IPAR, RPAR, IERR)
	   CALL FDIF(M, T, 0, F, IPAR, RPAR, IERR)
	   NEQ = M
	END IF
C
C   Setup right hand side
	CALL DSCAL(M, H, F, 1)
	CALL DLASET('A', N, 1, 0.0D0, 0.0D0, RWORK, N)
	DO 30 IO = 1, IORD-1
	   DO 30 I = 1,N
            RWORK(I) = RWORK(I)-COEFF(IO,IORD)*XQ((IORD-IO-1)*N+I)
  30  CONTINUE
	CALL DGEMV('N', NEQ, N, 1.0D0, E, M, RWORK, 1, 1.0D0, F, 1)
      DO 40 I = 1, NEQ
	   DO 40 J = 1, N
	      E(I,J) = COEFF(0,IORD)*E(I,J) - H*A(I,J)
  40  CONTINUE
C
C   Compute the least squares solution at t=t_0+h.
	CALL DGELS('N', NEQ, N, 1, E, M, F, MAX(M,N), RWORK, LRWORK, IERR)
	IF (IERR .LT. 0) RETURN
      CALL DCOPY(N, F, 1, XQ((IORD-1)*N+1), 1)
	DO 100 ISTEP = IORD+1, NSTEP
	   T = ISTEP*H
C
C   Compute the coefficients of the stangeness free system.
	   IF (IRED .EQ. 0) THEN
	      CALL REDUCE(EDIF, ADIF, FDIF, M, N, T, MAXIND, CVAL, E, M,
     $                  A, M, F, IPAR, RPAR, RWORK, LRWORK, IERR)
	   ELSE
	      CALL EDIF(M, N, T, 0, E, M, IPAR, RPAR, IERR)
	      CALL ADIF(M, N, T, 0, A, M, IPAR, RPAR, IERR)
	      CALL FDIF(M, T, 0, F, IPAR, RPAR, IERR)
	   END IF
         CALL DLASET('A', N, 1, 0.0D0, 0.0D0, RWORK, N)
	   DO 50 IO = 1, IORD
	      DO 50 I=1,N
	         RWORK(I) = RWORK(I)-COEFF(IO,IORD)*XQ((ISTEP-IO-1)*N+I)
  50     CONTINUE
         CALL DGEMV('N', NEQ, N, 1.0D0, E, M, RWORK, 1, H, F, 1)
         DO 60 I = 1, NEQ
	      DO 60 J = 1, N
	         E(I,J) = COEFF(0,IORD)*E(I,J) - H*A(I,J)
  60     CONTINUE
	   CALL DGELS('N', NEQ, N, 1, E, M, F, MAX(M,N), RWORK, LRWORK,
     $              IERR)
	   IF (IERR .LT. 0) RETURN
         CALL DCOPY(N, F, 1, XQ((ISTEP-1)*N+1), 1)
 100  CONTINUE
      RETURN
	END