      SUBROUTINE DNFOTRM(M, N, MQ, NQ, ID, IA, IU, IV, IREQ, EQ, LDEQ, 
     $                  AQ, LDAQ, Z2Q, LDZ2Q, Z1, LDZ1, AH, LDAH, RWORK, 
     $                  LRWORK, IERR)
C
C     PURPOSE
C
C     DNFOTRM is a slightly modified version of the subroutine DNFOTR,
C     which is part of the code DGELDA and has been adapted to
C     rectangular systems.
C
C     DNFOTRM computes the (local) orthogonal transoformations which
C     extract a strangeness free DAE from the extended derivative array
C     at a fixed time.
C
C     ARGUMENT LIST
C
C         M - INTEGER.
C             The number of equations in the DAE system.
C             M .GE. 1.
C
C         N - INTEGER.
C             The number of variables in the DAE system.
C             N .GE. 1.
C
C        NQ - INTEGER.
C             The size of EQ and AQ, i.e. NQ=(MU+1)*N, where MU is the
C             strangeness index of the DAE.
C             NQ .GE. 1.
C
C        ID - INTEGER.
C             The number DMU of differential components.
C
C        IA - INTEGER.
C             The number AMU of algebraic components.
C
C        IU - INTEGER.
C             The number UAU of undetermined components.
C
C      IREQ - INTEGER.
C             The rank of the matrix EQ.
C
C        EQ - DOUBLE PRECISION array of DIMENSION (LDEQ,*).
C             The leading NQ by NQ part of this array contains the
C             matrix EQ.
C
C      LDEQ - INTEGER.
C             The leading dimension of array EQ as declared in the
C             calling program.
C             LDEQ .GE. NQ.
C
C        AQ - DOUBLE PRECISION array of DIMENSION (LDAQ,*).
C             The leading NQ by NQ part of this array contains the
C             matrix AQ.
C
C      LDAQ - INTEGER.
C             The leading dimension of array AQ as declared in the
C             calling program.
C             LDAQ .GE. NQ.
C
C       Z2Q - DOUBLE PRECISION array of DIMENSION (LDZ2Q,*)
C             The leading NQ by NQ part of this array must contain the
C             right orthogonal factor of the SVD EQ=ZQ*S*VT.
C             NOTE that this array is overwritten.
C
C     LDZ2Q - INTEGER.
C             The leading dimension of array Z2Q as declared in the
C             calling program.
C             LDZ2Q .GE. NQ.
C
C      LDZ1 - INTEGER.
C             The leading dimension of array Z1 as declared in the
C             calling program.
C             LDZ1 .GE. N.
C
C      LDAH - INTEGER.
C             The leading dimension of array AH as declared in the
C             calling program.
C             LDZ1 .GE. NQ.
C
C       ARGUMENTS OUT
C
C       Z2Q - DOUBLE PRECISION array of DIMENSION (LDZ2Q,*).
C             The leading NQ by IA part of this array contains the
C             transformation matrix Z2Q which extracts the algebraic
C             part.
C
C        Z1 - DOUBLE PRECISION array of DIMENSION (LDZ1,*).
C             The leading N by ID part of this array contains the
C             transformation matrix Z1 which extracts the differential
C             part.
C
C        AH - DOUBLE PRECISION array of DIMENSION (LDAH,*).
C             This array is used as workarray.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - NTEGER.
C              The length of RWORK.
C              LRWORK .GE. 3*N + 2*NQ + 2*NQ*NQ.
C
C              NOTE that for good performance, LRWORK should generally
C              be larger.
C
C     ERROR INDICATOR
C
C       IERR - INTEGER.
C              Unless the routine detects an error (see next section),
C              IERR contains 0 on exit.
C
C     WARNINGS AND ERRORS DETECTED BY THE ROUTINE
C
C     IERR = -3 : An argument of DGESVD had an illegal value.
C     IERR = -4 : DGESVD failed to converge.
C     IERR = -5 : LRWORK < 3*N + 2*NQ + 2*NQ*NQ.
C
C     METHOD
C
C     All information we need to transform the original DAE is hidden in
C     the extended derivative array EQ, AQ anf FQ. The two orthogonal
C     transformation matricies Z1 and Z2Q are computed via two singular
C     value decompositions (SVDs). The LAPACK subroutine DGESVD is used
C     to compute these SVDs.
C
C
      IMPLICIT NONE
C     .. Parameters ..
      DOUBLE PRECISION ZERO, ONE
      PARAMETER        (ZERO = 0.0D0, ONE = 1.0D0)
      INTEGER          OSIGMA
      PARAMETER        (OSIGMA = 1)
C     .. Scalar Arguments ..
      INTEGER          IA, ID, IERR, IREQ, LDEQ, LDAH, LDAQ, LDZ1,
     $                 LDZ2Q, LRWORK, N, NQ, M, MQ, IU, IV,i,j
C     .. Array Arguments ..
      DOUBLE PRECISION EQ(LDEQ,*), AH(LDAH,*), AQ(LDAQ,*), RWORK(*),
     $                 Z1(LDZ1,*), Z2Q(LDZ2Q,*)
C     .. Local Scalars
      INTEGER          INFO, LWORK, OU, OVT, OWORK, NMIN
C     .. External Subroutines ..
      EXTERNAL         DGEMM, DGESVD, DLACPY, DLASET
C     .. Executable Statements ..
C
C     SET UP RWORK STORAGE.
      NMIN = MIN0(MQ,NQ) 
      OU    = OSIGMA + NMIN
      OVT   = OU + MQ*MQ
      OWORK = OVT + NQ*NQ
      LWORK = LRWORK - 3*N - 2*NQ - NQ*NQ - MQ*MQ
      IF (LWORK.LT.0) THEN
        IERR = 10
	RETURN
      ENDIF
C
C     NOTE: IREQ .EQ. MQ, iff the DAE is an ODE, i.e. MU=0 and NQ .EQ. N
      IF (IREQ.EQ.NQ) THEN
         CALL DLASET('N', MQ, MQ, ZERO, ONE, Z2Q, LDZ2Q)
         CALL DLASET('N', M, M, ZERO, ONE, Z1, LDZ1)
         RETURN
      END IF
C
C     Compute AH = Z2Q**T * AQ.
      CALL DGEMM('T','N', IA+IV, N, MQ, ONE, Z2Q, LDZ2Q, AQ, LDAQ,
     $           ZERO, AH, LDAH)
      CALL DGESVD('A','A', IA+IV, N, AH, LDAH, RWORK(OSIGMA), RWORK(OU),
     $            MQ, RWORK(OVT), NQ, RWORK(OWORK), LWORK, INFO)
C
C     Compute AH = Z2Q * U.
      CALL DGEMM('N','N', MQ, IA+IV, IA+IV, ONE, Z2Q, LDZ2Q, RWORK(OU),
     $           MQ, ZERO, AH, LDAH)
C
C     Set Z2Q = AH.
      CALL DLACPY('F', MQ, IA+IV, AH, LDAH, Z2Q, LDZ2Q)
      IF (ID.NE.0) THEN
C
C     Compute Z1 = EQ * VT**T. NOTE: Z1 is NxN-IA <= NxN.
         CALL DGEMM('N','T', M, N-IA, N, ONE, EQ, LDEQ, RWORK(OVT+IA),
     $              NQ, ZERO, Z1, LDZ1)
         CALL DGESVD('O','N', M, N-IA, Z1, LDZ1, RWORK(OSIGMA),
     $               RWORK(OU), NQ, RWORK(OVT), NQ, RWORK(OWORK), LWORK,
     $               INFO)
      END IF
      RETURN
C *** Last line of DNFOTR ***
      END
