      SUBROUTINE DNFVALM(M, N, MQ, NQ, EQ, LDEQ, AQ, LDAQ, R0,  A0, S0,
     $                   D0, U0, V0, ZQ, LDZQ, RWORK, LRWORK, IERR)
C
C     PURPOSE
C
C     DNFVALM is a slightly modified version of the subroutine DNFVAL,
C     which is part of the code DGELDA and has been adapted to
C     rectangular systems.
C
C     DNFVALM computes the characteristic values of a matrix pair (EQ,AQ)
C     and the right orthogonal factor ZQ of the SVD EQ = ZQ*S*VT.
C
C     ARGUMENT LIST
C
C       ARGUMENTS IN
C
C         M - INTEGER.
C             The number of equations in the DAE system.
C             M .GE. 1.
C
C         N - INTEGER.
C             The number of variables in the DAE system.
C             N .GE. 1.
C
C        NQ - INTEGER.
C             The size of the square matricies EQ and AQ.
C             NQ .GE. N.
C
C        EQ - DOUBLE PRECISION array of DIMENSION (LDEQ,*).
C             The leading NQ by NQ part of this array must contain the
C             matrix EQ.
C
C      LDEQ - INTEGER.
C             The leading dimension of array EQ as declared in the
C             calling program.
C             LDEQ .GE. NQ.
C
C        AQ - DOUBLE PRECISION array of DIMENSION (LDAQ,*).
C             The leading NQ by NQ part of this array must contain the
C             matrix AQ.
C
C      LDAQ - INTEGER.
C             The leading dimension of array AQ as declared in the
C             calling program.
C             LDAQ .GE. NQ.
C
C      LDZQ - INTEGER.
C             The leading dimension of array ZQ as declared in the
C             calling program.
C             LDAQ .GE. NQ.
C
C       ARGUMENTS OUT
C
C        R0 - INTEGER.
C             The rank of EQ.
C
C        A0 - INTEGER.
C             The size of the algebraic part.
C
C        S0 - INTEGER.
C             The size of the strangeness part.
C
C        D0 - INTEGER.
C             The size of the differential part.
C
C        U0 - INTEGER.
C             The size of the undetermind part.
C
C        ZQ - DOUBLE PRECISION array of DIMENSION (LDZQ,*).
C             The leading NQ by NQ part of this array contains the right
C             orthogonal factor of the SVD EQ = ZQ*S*VT.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - NTEGER.
C              The length  of RWORK.
C              LRWORK .GE. 3*NQ*NQ + 6*NQ.
C
C              NOTE that for good performance, LRWORK should generally
C              be larger.
C
C     ERROR INDICATOR
C
C       IERR - INTEGER.
C              Unless the routine detects an error (see next section),
C              IERR contains 0 on exit.
C
C     WARNINGS AND ERRORS DETECTED BY THE ROUTINE
C
C     IERR = -3 : An argument of DGESVD had an illegal value.
C     IERR = -4 : DGESVD failed to converge.
C     IERR = -5 : LRWORK < 3NQ*NQ + 6*NQ.
C
C     METHOD
C
C     The characteristic values of a NQ by NQ matrix pair (EQ,AQ) are
C     defined by
C
C                R0 = rank EQ
C                A0 = rank (Z^* AQ T)
C                S0 = rank(V^* Z^* AQ T')
C                D0 = R0 - S0
C                U0 = NQ - R0 - A0 - S0
C
C     where
C
C                T basis of kernel EQ
C                Z basis of corange EQ
C                T' basis of cokernel EQ
C                V basis of corange(Z^* AQ T).
C
C     The characteristic values R0,A0,S0,D0,U0 are computed via three
C     rank decisions, which are obtained by means of singular value
C     decompositions (SVDs). The LAPACK subroutine DGESVD is used to
C     compute these SVDs.
C
C
C     .. Parameters ..
      DOUBLE PRECISION ZERO, ONE
      PARAMETER        (ZERO = 0.0D0, ONE = 1.0D0)
      INTEGER          OWQ
      PARAMETER        (OWQ = 1)
C     .. Scalar Arguments ..
      INTEGER          A0, D0, IERR, LDEQ, LDAQ, LDZQ, LRWORK, M, NQ,
     $                 R0, S0, U0, V0
C     .. Array Arguments ..
      DOUBLE PRECISION AQ(LDAQ,*), EQ(LDEQ,*), RWORK(*), ZQ(LDZQ,*)
C     .. Local Scalars ..
      DOUBLE PRECISION UMACH
      INTEGER          I, LWORK, OS, OU, OVT, OWORK
C     .. External Subroutines ..
      EXTERNAL         DGEMM, DGESVD, DLACPY
C     .. Exectutable Statements ..
      NMAX = MAX0(MQ,NQ)
      NMIN = MIN0(MQ,NQ)
C
C     Set up rwork storage.
C
      OS    = OWQ + NMAX*NMAX
      OU    = OS + NMIN
      OVT   = OU + NMAX*NMAX
      OWORK = OVT + NQ*NQ
      LWORK = LRWORK - 3*NQ*NQ - 6*NQ
      IF (LWORK.LT.0) THEN
        IERR = -5
	RETURN
      ENDIF
      UMACH=1.D-12
C
C     Determine R0.
      CALL DLACPY('N', MQ, NQ, EQ, LDEQ, RWORK(OWQ), MQ)
      CALL DGESVD('A','A', MQ, NQ, RWORK(OWQ), MQ, RWORK(OS), ZQ, LDZQ,
     $            RWORK(OVT), NQ, RWORK(OWORK), LWORK, IERR)
      IF (IERR.NE.0) THEN
         IF (IERR.LT.0) THEN
            IERR = -3
         ELSE
            IERR = -4
         ENDIF
         RETURN
      ENDIF
      MIN = MIN0(MQ,NQ)
      R0=0
      IF (RWORK(OS) .GT. UMACH) THEN
         DO 10 I=0,MIN-1
            IF (RWORK(OS+I) .GT. UMACH*RWORK(OS)) THEN
               R0=R0+1
            END IF
 10      CONTINUE
      END IF
C
C     Determine A0.
      NAV = MQ - R0
      NAU = NQ - R0
      IF (NAV.NE.0 .AND. NAU.NE.0) THEN
C
C       WQ(NAV,NAU)=U(1:MQ,R0+1:MQ)^T AQ(1:MQ,1:N) VT(R0+1:NQ,1:N)^T
        CALL DGEMM('T','N', NAV, N, MQ, ONE, ZQ(1,R0+1), LDZQ, AQ, LDAQ,
     $             ZERO, RWORK(OU), NMAX)
	  CALL DGEMM('N','T', NAV, NAU, N, ONE, RWORK(OU), NMAX,
     $             RWORK(OVT+R0), NQ, ZERO, RWORK(OWQ), NMAX)
        CALL DGESVD('A','N', NAV, NAU, RWORK(OWQ), NMAX, RWORK(OS),
     $              RWORK(OU), MQ, RWORK(OVT), NQ, RWORK(OWORK), LWORK,
     $              IERR)
        IF (IERR.NE.0) THEN
           IF (IERR.LT.0) THEN
              IERR = -3
           ELSE
              IERR = -4
           ENDIF
           RETURN
        ENDIF
        MIN = MIN0(NAV,NAU)
        A0 = 0
        IF (RWORK(OS) .GT. UMACH) THEN
          DO 20 I=0,MIN-1
            IF (RWORK(OS+I) .GT. UMACH*RWORK(OS)) THEN
              A0=A0+1
            END IF
 20       CONTINUE
        END IF
      ELSE
        A0 = 0
      END IF
C
C     Determine S0.
      NSV = MQ - R0 - A0
      NU  = NQ - R0 - A0
      IF (NSV.NE.0 .AND. NU.NE.0 .AND. R0.NE.0) THEN
C
C        WQ(1:S0,1:R0)=UQ(1:NQ-R0,A0+1:A0+S0)^T U(1:NQ,R0+1:R0+NQ)^T
C                     *AQ(1:NQ,1:N) VT(1:R0,1:N)^T
         CALL DGEMM('T','T', NSV, MQ, NAV, ONE, RWORK(OU+A0*MQ), MQ,
     $              ZQ(1,R0+1), LDZQ, ZERO, RWORK(OWQ), NMAX)
         CALL DGEMM('N','N', NSV, N, MQ, ONE, RWORK(OWQ), NMAX,
     $              AQ, LDAQ, ZERO, RWORK(OU), MQ)
	   CALL DGEMM('N','T', NSV, R0, N, ONE, RWORK(OU), MQ, RWORK(OVT),
     $              NQ, ZERO, RWORK(OWQ), NMAX)
         CALL DGESVD('N','N', NSV, R0, RWORK(OWQ), NMAX, RWORK(OS),
     $               RWORK(OU), NQ, RWORK(OVT), NQ, RWORK(OWORK), LWORK,
     $               IERR)
         IF (IERR.NE.0) THEN
            IF (IERR.LT.0) THEN
               IERR = -3
            ELSE
               IERR = -4
            ENDIF
            RETURN
         ENDIF
         MIN=MIN0(NSV,NU)
         S0=0
         IF (RWORK(OS) .GT. UMACH) THEN
            DO 30 I=0,MIN-1
               IF (RWORK(OS+I) .GT. UMACH*RWORK(OS)) THEN
                  S0=S0+1
               END IF
 30         CONTINUE
         END IF
      ELSE
        S0 = 0
      END IF
C
C     Finish computation and print results.
      D0 = R0 - S0
      U0 = NQ - R0 - A0
      V0 = MQ - R0 - A0 - S0
      RETURN
C *** Last line of DNFVAL ***
      END
