      SUBROUTINE REDUCE(EDIF, ADIF, FDIF, M, N, T, MAXIND, CVAL, E, LDE,
     $                  A, LDA, F, IPAR, RPAR, RWORK, LRWORK, IERR)
C
C     PURPOSE
C
C     REDUCE performs the following steps:
C
C     1. Compute the strangeness index MU and the other characteristic
C        values of the matrix pair (E(t), A(t)) and time T.
C     2. Compute the matricies EQ, AQ und the vector FQ of the extended
C        derivative array of stage MU at time T.
C     3. Check if the strangeness index and the other characteristic values
C        have changed.
C     4. Compute the coefficients of an equivalent strangeness free DAE
C        at time T. The two DAEs are equivalent in the sense that their
C        solutions are identical.
C
C     REDUCE is a slightly modified version of the subroutine DNFIX1, 
C     which is a part of the software package DGELDA. It has been adapted
C     such that it allows for rectangular systems and the parameter list
C     has been simplified.
C
C     ARGUMENT LIST
C
C        USER-SUPPLIED SUBROUTINES
C
C      EDIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             matrix E(t) and its derivatives. It has the form
C
C             SUBROUTINE EDIF(N,T,IDIF,E,LDE,IPAR,RPAR,IERR).
C
C             The subroutine takes as input the number of equations N,
C             the time T and the integer parameter IDIF. EDIF must not
C             alter these input parameters or the leading dimension LDE
C             of the array E. As output, the subroutine produces the
C             IDIF-th derivative of E(t) at time T in the leading N by N
C             part of the array E. The integer flag IERR is always zero
C             on input and EDIF should alter IERR only if IDIF is larger
C             than the highest derivative of E(t) the subroutine
C             provides (set IERR=-2) or if another problem occurs (set
C             IERR=-1). IPAR and RPAR are integer and real arrays which
C             can be used for the communication between the user's
C             calling program and the subroutines EDIF, ADIF and FDIF.
C
C             In the calling program, EDIF must be declared as external.
C
C      ADIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             matrix A(t) and its derivatives. It is of the form
C
C              SUBROUTINE ADIF(N,T,IDIF,A,LDA,IPAR,RPAR,IERR)
C
C             and the input and output parameters are similar to these
C             of EDIF.
C
C      FDIF - User supplied SUBROUTINE.
C             This is a subroutine which the user provides to define the
C             vector f(t) and its derivatives. It is of the form
C
C             SUBROUTINE FDIF(N,T,IDIF,F,IPAR,RPAR,IERR)
C
C             and the input and output parameters are similar to these
C             of EDIF. Except that the first N elements of the
C             1-dimensional array F contain the IDIF-th derivative of
C             f(t) at time T. Note further, since F is a 1-dimensional
C             array no leading dimension is needed.
C
C        ARGUMENTS IN
C
C         M - INTEGER.
C             The number of equations in the DAE system.
C             M .GE. 1.
C
C         N - INTEGER.
C             The number of variables in the DAE system.
C             N .GE. 1.
C
C         T - DOUBLE PRECISION.
C             The time T.
C
C    MAXIND - INTEGER.
C             The maximal strangeness index.
C             MAXIND .GE. 1.
C
C       LDE - INTEGER.
C             The leading dimension of array E as declared in the
C             calling program.
C             LDE .GE. N.
C
C       LDA - INTEGER.
C             The leading dimension of array A as declared in the
C             calling program.
C             LDA .GE. N.
C
C     ARGUMENTS OUT
C
C      CVAL - INTEGER array of DIMENSION (5)
C             If INFO(1)=0 this array contains the characteristic
C             values of the DAE
C             CVAL(1) contains the strangeness index MU.
C             CVAL(2) contains the number DMU of differential
C                     components.
C             CVAL(3) contains the number AMU of algebraic components.
C             CVAL(4) contains the number UMU of undetermined
C                     components.
C             CVAL(5) contains the number VMU of solvability
C                     conditions.
C
C         E - DOUBLE PRECISION array of DIMENSION (LDE,*).
C             The leading N by N part of this array contains the matrix
C             E of the strangeness free DAE at time T.
C
C         A - DOUBLE PRECISION array of DIMENSION (LDA,*).
C             The leading N by N part of this array contains the matrix
C             A of the strangeness free DAE at time T.
C
C         F - DOUBLE PRECISION array of DIMENSION (*).
C             The leading N elements of this array contain the vector F of
C             the strangeness free DAE at time T.
C
C      IPAR - INTEGER array of DIMENSION (*).
C             This integer array can be used for communication between
C             the calling program and the EDIF, ADIF and FDIF
C             subroutines.
C
C      RPAR - DOUBLE PRECISION array of DIMENSION (*).
C             This real array can be used for communication between the
C             calling program and the EDIF, ADIF and FDIF subroutines.
C
C             IPAR and RPAR are not altered by DNFPEN. If these arrays
C             are used, they must be dimensioned in the calling program
C             and in EDIF, ADIF and FDIF as arrays of appropriate
C             length. Otherwise, ignore these arrays by treating them as
C             dummy arrays of length one.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - NTEGER.
C              The length of RWORK.
C              LRWORK .GE. 3*((MAXIND+1)*N)^2 + 6*(MAXIND+1)*N.
C
C              NOTE that for good performance, LRWORK should generally
C              be larger.
C
C     ERROR INDICATOR
C
C       IERR - INTEGER.
C              Unless the routine detects an error (see next section),
C              IERR contains 0 on exit.
C
C     WARNINGS AND ERRORS DETECTED BY THE ROUTINE
C
C     IERR = -1  : An error occured in EDIF, ADIF or FDIF.
C     IERR = -2  : The input parameter IDIF of EDIF, ADIF or FDIF was
C                  larger then the highest derivative the routine
C                  provides.
C     IERR = -22 : Failed to determine strangeness index.
C     IERR = -23 : Failed to compute an equivalent strangeness index 0
C                  system.
C     IERR = -26 : Unable to continue due to a change in characteristic
C                  values.
C
C     METHOD
C
C     If we denote the i-th derivative by a superscript (i) then EQ, AQ
C     and FQ are defined componentwise by
C
C                  /i\   (i-j)      / i \   (i-j-1)
C      (EQ)    := |   | E    (T) - |     | A      (T)  ; i,j = 0,...,L
C          i,j     \j/              \j+1/
C
C                  /   (i)
C                 |   A   (T) ; for i = 0,...,L , j=0
C      (AQ)    := |
C          i,j     \  0       ; otherwise
C
C                  (i)
C      (FQ)    := f  (T)      ; i = 0,...,L
C          i
C
C     The characteristic values of the matrix pair (EQ,AQ) of stage I are
C     defined by
C
C                RI = rank EQ
C                AI = rank (Z^* AQ T)
C                SI = rank(V^* Z^* AQ T')
C                DI = RI - SI
C                UI = N*(I+1) - RI - AI - SI
C
C     where
C
C                T basis of kernel EQ
C                Z basis of corange EQ
C                T' basis of cokernel EQ
C                V basis of corange(Z^* AQ T).
C
C     Recursive formulas are used to compute the strangeness index MU
C     and the characteristics of the original pair (E(t), A(t)) at time
C     T in terms of the above values.
C
C     All information we need to compute the strangeness free DAE is
C     hidden in the extended derivative array EQ, AQ anf FQ.
C
C     .. Parameters ..
      DOUBLE PRECISION ZERO, ONE
      PARAMETER        (ZERO = 0.0D0, ONE = 1.0D0)
C     .. Subroutine Arguments ..
      EXTERNAL         EDIF, ADIF, FDIF
C     .. Scalar Arguments ..
      DOUBLE PRECISION T
      INTEGER          IERR, ID, IA, IU, IDQ, IAQ, IUQ, IREQ,
     $                 LDA, LDAH, LDAQ, LDE, LDEQ, LDZ1, LDZ2Q,
     $                 MAXIND, LRWORK, MU, MQ, M, N
      LOGICAL          MCONST
C     .. Array Arguments ..
      DOUBLE PRECISION A(LDA,*), E(LDE,*), F(*), RPAR(*), RWORK(*)
      INTEGER          CVAL(*), IPAR(*)
C     .. Local Scalars ..
      INTEGER          I, NQ, OEQ, OAQ, OAH, OFQ, OZ2Q, OZ1, OWORK
C     .. External Subroutines ..
      EXTERNAL         DCOPY, DGEMV, DLACPY, DLASET
C     .. Executable Statements ..
      MQ = M*(MAXIND+1)
      NQ = N*(MAXIND+1)
C
C     Compute relevant matrices.
      IERR = 0
	LDEQ  = MQ
	LDAQ  = MQ
	LDAH  = MQ
	LDZ2Q = MQ
	LDZ1  = M
	OEQ   = 1
	OAQ   = OEQ + LDEQ*NQ
	OAH   = OAQ + LDAQ*NQ
	OFQ   = OAH + LDAH*NQ
	OZ2Q  = OFQ + MQ
	OZ1   = OZ2Q + LDZ2Q*MQ
	OWORK = OZ1 + LDZ1*N
	LWORK = LRWORK - OWORK
      CALL DNFINDM(EDIF, ADIF, FDIF, M, N, T, MAXIND, MU2, ID, IA, IU,
     $             IV, IREQ, RWORK(OAH), LDAH, RWORK(OEQ), LDEQ,
     $             RWORK(OAQ), LDAQ, RWORK(OFQ), RWORK(OZ2Q), LDZ2Q,
     $             IPAR, RPAR, RWORK(OWORK), LWORK, IERR)
      IF (IERR.NE.0) THEN
         IERR = -22
         RETURN
      ENDIF
      MQ = M*(MU2+1)
      NQ = N*(MU2+1)
      CALL DNFREDM(M, N, MQ, NQ, ID, IA, IU, IV, IREQ, RWORK(OEQ), LDEQ, 
     $ 	         RWORK(OAQ), LDAQ, RWORK(OFQ), RWORK(OZ2Q),
     $             LDZ2Q, RWORK(OZ1), LDZ1, E, LDE, A, LDA, F,
     $             RWORK(OAH), LDAH, RWORK(OWORK), LRWORK, IERR)
      IF (IERR.NE.0) THEN
         IERR = -23
         RETURN
      ENDIF
	CVAL(1) = MU2
	CVAL(2) = ID
	CVAL(3) = IA
	CVAL(4) = IU
	CVAL(5) = IV
      RETURN
C *** Last line of DNFIX1 ***
      END
