      SUBROUTINE SLVBVP(NEQNS, NSTEP, XQ, RWORK, LRWK)
C     .. Scalar Arguments..
      INTEGER          NEQNS, NSTEP, LRWK
C     .. Array Arguments ..
      DOUBLE PRECISION XQ(*), RWORK(*)
C
C  Purpose:
C   
C   Computes the solution of the minimization problem
C                            .
C      ||x||_2=min!  s.t. ||Ex-Ax-f||_2=min!,
C
C   where
C       .
C      Ex = Ax + f, x(t_0)=0 
C
C   is a strangeness-free linear differential algebraic equation in
C   orthogonal standard form. The solution of this minimization problem
C   corresponds with the solution of a boundary value problem of the
C   form
C
C     (d/dt)[y_1]=[   A     BB^T ][y_1]+[ f  ],
C           [y_2] [ I+C^TC  A^T  ][y_2]+[C^Tg]
C
C     y_1(t_0)=0, y_2(T)=0,
C
C   which is computed using the code MIRKDC.
C
C
C       ARGUMENTS IN
C
C
C     NEQNS - INTEGER.
C             The number of equations in the ODE system. 
C             N .GE. 1.
C
C     NSTEP - INTEGER
C             The number of steps where the solution evaluated.
C             NSTEP .GE. 1.
C
C
C       ARGUMENTS OUT
C
C        XQ - DOUBLE PRECISION array of DIMENSION (NSTEP*N)
C             For i=1:NSTEP XQ((i-1)+1:i*N) contains the 
C             approximation x_i to the solution of the boundary value
C             problem at the timestep t_i=t_0+i*h.
C
C     WORK SPACE
C
C      RWORK - DOUBLE PRECISION array of DIMENSION at least (LRWORK).
C
C     LRWORK - NTEGER.
C              The length of RWORK.
C              LRWORK .GE. 1000*NEQNS*(2*NEQNS+35) + 6*NEQNS*(5*NEQNS+2).
C
C
C     The following subroutines have to be provided by the user:
C
C    SETSIZE - User supplied SUBROUTINE
C              This subroutine defines the size of the DAE system and
C              the characteristic values \hat d, \hat a, \hat u and
C              \hat v. It has the form
C                SUBROUTINE SETSIZE(M, N, ID, IA, IU, IV).
C              As output, this subroutine produces the number M of 
C              equations and the number N of variables of the 
C              strangeness-free DAE system. Furthermore it defines 
C              the characteristic values ID=\hat d, IA=\hat a, 
C              IU=\hat u and IV=\hat v of the DAE system.
C
C     SETINT - User supplied SUBROUTINE
C              This subroutine defines the left and right bounds of
C              the integration interval [t_0, T]. It has the form
C                SUBROUTINE SETINT(T0, TN).
C              As output, this subroutine defines the left and the 
C              right interval bounds T0=t_0 and TN=T.
C       
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C              If IPAR(1)=0, then the matrix E has to be defined 
C              according to the orthogonal standard form, i.e. the 
C              leading ID by ID part of E(t) has to be nonsingular,
C              the rest of the matrix has to be equal to zero.
C              If IPAR(1)=1, then the array E contains the matrix
C              E(t)*Q(t) where the orthogonal matrix Q(t) is defined by
C              the subroutine QDEF.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C              If IPAR(1)=0, then the matrix A has to be defined 
C              according to the orthogonal standard form, i.e. 
C              only the blocks A(1:ID,1:N) and A(ID+1:ID+IA,1:ID+IA)
C              contain nonzero entries and the block 
C              A(ID+1,ID+IA,ID+1:ID+IA) is nonsingular.
C              If IPAR(1)=1, then the array A contains the matrix
C              A(t)*Q(t)-E(t)*(d/dt)Q(t) where the orthogonal matrix 
C              Q(t) and its derivative (d/dt)Q(t) are defined by
C              the subroutines QDEF and QDOT.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C
C
C     .. Parameters ..
      INTEGER          MXNSUB
      PARAMETER        (MXNSUB = 1000)
C     .. Local Scalars ..
      INTEGER          METHOD, MESH_IN, SOL_IN, OUT_CNTR, INFO, NSUB,
     $                 M, N, ID, IA, IU, IV, OA, OB, OC, OF, OG,
     $                 FAIL, LENIWK, ORWORK, LENRWK
      DOUBLE PRECISION TOL, HEVAL, T, UEVAL
C     .. Local Arrays ..
      INTEGER          IWORK(NEQNS*(MXNSUB+1))
      DOUBLE PRECISION MESH(0:MXNSUB),Y(NEQNS,MXNSUB+1), YEVAL(NEQNS),
     $                 YPRIME(NEQNS)
C     .. External Subroutines ..
      EXTERNAL         INIT_DE, INIT_Y, FSUB, GSUB, DFSUB, DGSUB
C     .. Executable Statements ..
C
C  Get system properties
C
      CALL SETSIZE(M, N, ID, IA, IU, IV)
      IF (ID .EQ. 0) THEN
         WRITE (*,*) 'testbvp: ID=0, BVP-solver is not suitable here'
	   GOTO 1000
      END IF
C
C  Setup workspace
C
      OA     = 1
      OB     = OA + ID*ID
      OC     = OB + ID*IU
      OF     = OC + IA*ID
      OG     = OF + ID
	ORWORK = OG + IA
      LENRWK = MXNSUB*NEQNS*(2*NEQNS+35) + 6*NEQNS*(4*NEQNS+1)
      IF (LRWK-ORWORK .LT. LENRWK) THEN
         WRITE (6,100) LRWK, LENRWK+ORWORK
	   GOTO 1000
      END IF 
  100 FORMAT (1X,'SLVBVP: LRWK=',i8,
     $        ' is too small! Set LRWK at least to',i8)      
C
C  Set parameters for MIRKDC
C
C   Order of the MIRK-scheme      
      METHOD = 6
C   Tolerance      
      TOL = 1.0D-10
C   Number of initial meshpoints
      NSUB = 10    
C   Uniform initial mesh      
      MESH_IN = 0
C   Initial solution given in init_Y      
      SOL_IN = 1
C   Output control
      OUT_CNTR = 0 
C
C   Call MIRKDC
C
      CALL MIRKDC(METHOD, TOL, NEQNS, NSUB, MXNSUB, MESH_IN, MESH, 
     $            SOL_IN, Y, NEQNS, OUT_CNTR, IERR, IWORK,
     $            RWORK(ORWORK), INIT_DE, INIT_Y, FSUB, GSUB, DFSUB, 
     $            DGSUB)
      IF (IERR .NE. 0) THEN
         WRITE (*,*) 'testbvp: Solution of boundary value problem', 
     $               ' failed'
         GOTO 1000
      END IF
C
C   Evaluate the result at neval+1 points      
C
      HEVAL = 1.0D0/NSTEP
      FAIL = 0
      DO 10 I=1,NSTEP
         T=I*HEVAL
	   CALL SOL_EVAL(T, YEVAL, YPRIME, FAIL, IWORK, RWORK(ORWORK))
	   CALL DCOPY(ID, YEVAL, 1, XQ((I-1)*N+1), 1)
         CALL CALCOEFF(M, N, ID, IA, IU, IV, T, RWORK(OA), RWORK(OB), 
     $                 RWORK(OC), RWORK(OF), RWORK(OG))
	   IF (IA.GT.0) THEN
	      CALL DGEMV('N',IA, ID, -1.0D0, RWORK(OC), IA, YEVAL, 1,
     $                 -1.0D0, RWORK(OG), 1)
            CALL DCOPY(IA, RWORK(OG), 1, XQ((I-1)*N+ID+1), 1)
         END IF	
	   IF (IU .GT. 0) THEN    	    
	      CALL DGEMV('T',ID, IU, 1.0D0, RWORK(OB), ID, YEVAL(ID+1), 1,
     $                 0.0D0, XQ((I-1)*N+ID+IA+1), 1)
         END IF
  10  CONTINUE
1000  RETURN
      END   
      
      SUBROUTINE INIT_DE(LEFTBC, A, B)
C     .. Scalar Arguments..
      INTEGER          LEFTBC
      DOUBLE PRECISION A, B
C     .. Local Scalars ..
      INTEGER          M, N, ID, IA, IU, IV
C     .. Executable Statements ..
C
C   Get the system properties
C
      CALL SETSIZE(M, N, ID, IA, IU, IV)
C
C   Set number of left boundary conditions      
C
      LEFTBC = ID
C
C   Set endpoints of the interval     
C
      CALL SETINT(A, B)
      RETURN
      END
      
      SUBROUTINE INIT_Y(NSUB, NEQNS, MESH, Y, LDY)
C     .. Scalar Arguments..
      INTEGER          NSUB, NEQNS, LDY
C     .. Array Arguments ..
      DOUBLE PRECISION MESH(0:NSUB),Y(LDY,*)
C     .. Local Scalars ..
      INTEGER          I,J
C     .. Executable Statements ..
C
C   Set the initial guess for the solution at meshpoints to zero
C
      DO 10 J=0,NSUB
         DO 10 I=1,NEQNS
            Y(I,J+1) = 0.0d0
 10   CONTINUE
      RETURN
      END
      
      SUBROUTINE FSUB(NEQNS, T, Y, F)
C     .. Scalar Arguments..
      INTEGER          NEQNS
	DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Y(*), F(*)
C
C   This subroutine defines the coefficients of the boundary value
C   problem
C
C     (d/dt)[y_1]=[   A     BB^T ][y_1]+[ f  ],
C           [y_2] [ I+C^TC  A^T  ][y_2]+[C^Tg]
C
C     y_1(t_0)=0, y_2(T)=0,
C
C   for MIRKDC.
C
C     .. Local Scalars ..
      INTEGER          M, N, ID, IA, IU, IV, OA, OB, OC, OF, OG,
     $                 OMATRIX
C     .. Local Arrays ..
      DOUBLE PRECISION RWORK(1000)
C     .. Executable Statements ..      
C
C   Get the system properties and compute A, B, C, f, g
C
      CALL SETSIZE(M, N, ID, IA, IU, IV)
      OA      = 1
      OB      = OA + ID*ID
      OC      = OB + ID*IU
      OF      = OC + IA*ID
      OG      = OF + ID
      OMATRIX = OG + IA
      CALL CALCOEFF(M, N, ID, IA, IU, IV, T, RWORK(OA), RWORK(OB), 
     $              RWORK(OC), RWORK(OF), RWORK(OG))
C
C  Setup first block row:
C    f_1 = A*y_1 + BB^T*y_2 + f
C
      CALL DGEMV('N',ID, ID, 1.0D0, RWORK(OA), ID, Y, 1, 1.0D0,
     $           RWORK(OF), 1)
      IF (IU .GT. 0) THEN
         CALL DGEMM('N','T',ID, ID, IU, 1.0D0, RWORK(OB), ID, 
     $              RWORK(OB), ID, 0.0D0, RWORK(OMATRIX), ID)
         CALL DGEMV('N',ID, ID, 1.0D0, RWORK(OMATRIX), ID, Y(ID+1), 1,
     $              1.0D0, RWORK(OF), 1)
      END IF
      CALL DCOPY(ID, RWORK(OF), 1, F, 1)
C
C  Setup second block row:
C    f_2 = (I+C^TC)*y_1 + A^T*y_2 + C^Tg
C
      IF (IU .GT. 0) THEN
         CALL DLASET('A', ID, ID, 0.0D0, 1.0D0, RWORK(OMATRIX), ID)
         IF (IA .GT. 0) THEN
            CALL DGEMM('T','N',ID, ID, IA, 1.0D0, RWORK(OC), IA, 
     $                 RWORK(OC), IA, 1.0D0, RWORK(OMATRIX), ID)	 
         END IF
         CALL DGEMV('N',ID, ID, 1.0D0, RWORK(OMATRIX), ID, Y, 1, 
     $               0.0D0, F(ID+1), 1) 
         CALL DGEMV('T',ID, ID, -1.0D0, RWORK(OA), ID, Y(ID+1), 1,
     &               1.0D0, F(ID+1), 1)
         IF (IA .GT. 0) THEN
            CALL DGEMV('T',IA, ID, 1.0D0, RWORK(OC), IA, RWORK(OG), 1, 
     $                 1.0D0, F(ID+1), 1)
         END IF
      END IF
      RETURN
      END
      
      SUBROUTINE DFSUB(NEQNS, T, Y, JJ)
C     .. Scalar Arguments..
      INTEGER          NEQNS 
	DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Y(*), JJ(NEQNS,*)
C
C   This subroutine defines the jacobian of the boundary value
C   problem
C
C     (d/dt)[y_1]=[   A     BB^T ][y_1]+[ f  ]
C           [y_2] [ I+C^TC  A^T  ][y_2]+[C^Tg]
C
C     y_1(t_0)=0, y_2(T)=0,
C
C   for MIRKDC.
C
C     .. Local Scalars ..
      INTEGER          M, N, ID, IA, IU, IV, OA, OB, OC, OF, OG
C     .. Local Arrays ..
      DOUBLE PRECISION RWORK(1000)
C     .. Executable Statements ..         
C
C   Get the system properties and compute A, B, C, f, g
C
      CALL SETSIZE(M, N, ID, IA, IU, IV)
      OA = 1
      OB = OA + ID*ID
      OC = OB + ID*IU
      OF = OC + IA*ID
      OG = OF + ID
      CALL CALCOEFF(M, N, ID, IA, IU, IV, T, RWORK(OA), RWORK(OB), 
     $              RWORK(OC), RWORK(OF), RWORK(OG))
C
C  JJ_11 = A
C
      CALL DLACPY('N',ID, ID, RWORK(OA), ID, JJ, NEQNS)
      IF (IU .GT. 0) THEN  
C	   
C  JJ_22 = -A^T 
C
         DO 10 I=1,ID
            DO 10 J=1,ID
	       JJ(ID+I,ID+J)= -JJ(J,I)
  10     CONTINUE	
C      
C  JJ_21 = I + C^T * C
C
         CALL DLASET('A', ID, ID, 0.0D0, 1.0D0, RWORK(OA), ID)
         IF (IA.GT.0) THEN
            CALL DGEMM('T','N',ID, ID, IA, 1.0D0, RWORK(OC), IA, 
     $                 RWORK(OC), IA, 1.0D0, RWORK(OA), ID)	 
         END IF
         CALL DLACPY('N',ID, ID, RWORK(OA), ID, JJ(ID+1,1), NEQNS)
C
C  JJ_12 = B * B^T  
C
         CALL DGEMM('N','T',ID, ID, IU, 1.0D0, RWORK(OB), ID, 
     $              RWORK(OB), ID, 0.0D0, JJ(1,ID+1), NEQNS) 
      END IF
      RETURN
      END
      
      SUBROUTINE GSUB(NEQNS, YA, YB, BC)
C     .. Scalar Arguments..
      INTEGER          NEQNS
C     .. Array Arguments ..
      DOUBLE PRECISION YA(*), YB(*), BC(*)
C     .. Local Scalars ..
      INTEGER          M, N, ID, IA, IU, IV
C     .. Executable Statements ..         
C      
C   Setup the boundary conditions y_1(t_0)=0, y_2(T)=0
C
      CALL SETSIZE(M, N, ID, IA, IU, IV)
      DO 10 I=1,ID
         BC(I) = YA(I)
	   IF (IU .GT. 0) THEN
	      BC(ID+I) = YB(ID+I)
	   END IF
  10  CONTINUE	 
      RETURN
      END


      
      SUBROUTINE DGSUB(NEQNS, YA, YB, DBC)
C     .. Scalar Arguments..
      INTEGER          NEQNS
C     .. Array Arguments ..
      DOUBLE PRECISION YA(*), YB(*), DBC(NEQNS,*)
C     .. Local Scalars ..
      INTEGER          M, N, ID, IA, IU, IV, IDIF
C     .. Executable Statements ..    
C           
C   Setup the Jacobian of the boundary conditions y_1(t_0)=0, y_2(T)=0
C
      CALL DLASET('A', NEQNS, NEQNS, 0.0D0, 1.0D0, DBC, NEQNS)
      RETURN
      END
      
      SUBROUTINE CALCOEFF(M, N, ID, IA, IU, IV, T, AOUT, BOUT, COUT,
     $                    FOUT, GOUT)
C     .. Scalar Arguments..
      INTEGER          M, N, ID, IA, IU, IV
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION AOUT(ID,*), BOUT(ID,*), COUT(IA,*), FOUT(*),
     $                 GOUT(*) 
C
C   This subroutine defines the matrices A, B and C and the vectors
C   f and g which define the boundary value problem
C
C     (d/dt)[y_1]=[   A     BB^T ][y_1]+[ f  ]
C           [y_2] [ I+C^TC  A^T  ][y_2]+[C^Tg]
C
C     y_1(t_0)=0, y_2(T)=0,
C
C   from the coefficients of the initial value problem
C
C     [ \Sigma_E 0 0 ]      [x_1] [ A_{11}  A_{12}  A_{13} ] [f_1]
C     [     0    0 0 ](d/dt)[x_2]=[ A_{21} \Sigma_A   0    ]+[f_2],
C     [     0    0 0 ]      [x_3] [   0        0      0    ] [f_3]
C
C     x_1(t_0)=0,
C
C   i.e.
C
C     A = \Sigma_E^{-1}*(A_{11} - A_{12}*\Sigma_A^{-1}*A_{12}),
C     B = \Sigma_E^{-1}*A_{13},
C     C = -\Sigma_A^{-1}*A_{21},
C     f = \Sigma_E^{-1}*(f_1 - A_{12}*\Sigma_A^{-1}*f_2),
C     g = -\Sigma_A^{-1}*f_2.
C
C   It also checks if the differential algebraic system is given in
C   the orthogonal standard form.
C
C     .. Local Scalars ..
      INTEGER          OS, OU, OV, OSE, OA11, OA12, OA13, OA21, OSA,
     $                 ORWRK, LRWRK, IERR, I, J, INFO
	DOUBLE PRECISION RDUMMY
C     .. Local Arrays ..
      DOUBLE PRECISION E(M,N), A(M,N), F(M), RWORK(1000)
      INTEGER          IPIV(N)
C     .. Executable Statements ..   
C 
C  Setup RWORK-array
C
      OS    = 1
      OU    = OS + ID
      OV    = OU + 1
      ORWRK = OV + 1
	LRWRK = 1000 - ORWRK
C
C  Compute given values and check if the system is given in the
C  orthogonal standard form
C
      IDIF = 0
      CALL EDEF(M, N, T, IDIF, E, M, 0, RDUMMY, IERR)
	CALL DGESVD('N', 'N', ID, ID, E, M, RWORK(OS), RWORK(OU), 1,
     $            RWORK(OV), 1, RWORK(ORWRK), LRWRK, IERR)
	IF (RWORK(OS+ID-1) .LE. 1.0D-12) THEN
	   WRITE (*,*) 'Error: rank of Sigma_E < d'
	   WRITE (*,*) 'Program terminated'
	   STOP
	END IF
      DO 10 I = 1,ID
	   DO 10 J = ID+1, N
	      IF (E(I,J) .NE. 0.0D0) THEN
	         WRITE (*,*) 'Error: E not given in standard form'
	         WRITE (*,*) 'Program terminated'
	         STOP
	      END IF
  10  CONTINUE
      DO 20 I = ID+1,M
	   DO 20 J = 1, N
	      IF (E(I,J) .NE. 0.0D0) THEN
	         WRITE (*,*) 'Error: E not given in standard form'
	         WRITE (*,*) 'Program terminated'
	         STOP
	      END IF
  20  CONTINUE
      CALL ADEF(M, N, T, IDIF, A, M, 0, RDUMMY, IERR)
	IF (IA .GT. 0) THEN
	   CALL DGESVD('N', 'N', IA, IA, A(ID+1,ID+1), M, RWORK(OS),
     $               RWORK(OU), 1, RWORK(OV), 1, RWORK(ORWRK), LRWRK, 
     $               IERR)
	   IF (RWORK(OS+IA-1) .LE. 1.0D-12) THEN
	      WRITE (*,*) 'Error: rank of Sigma_A < a'
	      WRITE (*,*) 'Program terminated'
	      STOP
	   END IF
	END IF
      DO 30 I = ID+1,ID+IA
	   DO 30 J = ID+IA+1, N
	      IF (A(I,J) .NE. 0.0D0) THEN
	         WRITE (*,*) 'Error: A not given in standard form'
	         WRITE (*,*) 'Program terminated'
	         STOP
	      END IF
  30  CONTINUE
      DO 40 I = ID+IA+1,M
	   DO 40 J = 1, N
	      IF (A(I,J) .NE. 0.0D0) THEN
	         WRITE (*,*) 'Error: A not given in standard form'
	         WRITE (*,*) 'Program terminated'
	         STOP
	      END IF
  40  CONTINUE
      CALL FDEF(M, T, IDIF, F, 0, RDUMMY, IERR)
C
C  Setup RWORK-array
C
      OSE  = 1
      OA11 = OSE  + ID*ID
      OA12 = OA11 + ID*ID
      OA13 = OA12 + ID*IA
      OA21 = OA13 + ID*IU
      OSA  = OA21 + IA*ID
      CALL DLACPY('N',ID, ID, E, M, RWORK(OSE), ID)
      CALL DLACPY('N',ID, ID, A, M, RWORK(OA11), ID)
      IF (IA .GT. 0) THEN
	   CALL DLACPY('N',ID, IA, A(1,ID+1), M, RWORK(OA12), ID)
	   CALL DLACPY('N',IA, ID, A(ID+1,1), M, RWORK(OA21), IA)
	   CALL DLACPY('N',IA, IA, A(ID+1,ID+1), M, RWORK(OSA), IA)
      END IF
      IF (IU .GT. 0) THEN 
         CALL DLACPY('N', ID, IU, A(1,ID+IA+1), M, RWORK(OA13), ID)
      END IF
      IF (IA .GT. 0) THEN
C
C  A_21 <- S_A^(-1)*A_21	 
C
	 CALL DGESV(IA, ID, RWORK(OSA), IA, IPIV(ID+1), RWORK(OA21),
     $              IA, INFO)
C
C  f_2 <- S_A^(-1)*f_2     
C
         CALL DGETRS('N',IA, 1, RWORK(OSA), IA, IPIV(ID+1), F(ID+1),
     $                  M, INFO)	    
         IF (INFO.NE.0) STOP
C
C  A_11 <- -A_12*A_21 + A_11	    
C
	 CALL DGEMM('N','N',ID, ID, IA, -1.0D0, RWORK(OA12), ID,
     $                 RWORK(OA21), IA, 1.0D0, RWORK(OA11), ID)
C
C  f_1 <- -A_12*f_2 + f_1     
C
         CALL DGEMV('N',ID, IA, -1.0D0, RWORK(OA12), ID, F(ID+1), 1,
     $                 1.0D0, F, 1)	    
      END IF
C
C  A_11 <- S_E^(-1)*A_11	 
C
      CALL DGESV(ID, ID, RWORK(OSE), ID, IPIV, RWORK(OA11), ID, 
     $              INFO)
C
C  f_1 <- S_E^(-1)*f_1     
C
      CALL DGETRS('N',ID, 1, RWORK(OSE), ID, IPIV, F, M, INFO)
C
C  A_11 -> A
C
      CALL DLACPY('N',ID, ID, RWORK(OA11), ID, AOUT, ID)
C
C  f_1 -> f      
C
      CALL DCOPY(ID, F, 1, FOUT, 1)
C
C  A_13 <- S_E^(-1)*A_13 
C
      IF (IU.GT.0) THEN     
         CALL DGETRS('N',ID, IU, RWORK(OSE), ID, IPIV, RWORK(OA13), 
     $               ID, INFO)
C
C  A_13 -> B     
C
         CALL DLACPY('N',ID, IU, RWORK(OA13), ID, BOUT, ID)
      END IF
      CALL FDEF(M, T, IDIF, F, 0, RDUMMY, IERR)
      IF (IA.GT.0) THEN
C
C  A_21 <- S_A^(-1)*A_21      
C
         CALL DGETRS('N',IA, ID, RWORK(OSA), IA, IPIV(ID+1), A(ID+1,1),
     $               M, INFO)
C
C  A_12 -> C     
C
         CALL DLACPY('N',IA, ID, A(ID+1,1), M, COUT, IA)
C
C  f_2 <- S_A^(-1)*f_2	 	  
C
         CALL DGETRS('N',IA, 1, RWORK(OSA), IA, IPIV(ID+1), F(ID+1), M, 
     $               INFO)
C
C  f_2 -> g     
C
         CALL DCOPY(IA, F(ID+1), 1, GOUT, 1)
      END IF
      RETURN
      END
