C=======================================================================
C     Driver routine for the subroutines DLOCAL and DGLOBAL
C=======================================================================
C
C   This program solves an over- or underdetermined linear 
C   strangeness-free DAE-system of the form
C         .
C     E(t)x(t) = A(t)x(t) + f(t)
C
C   by solving a discretization with the implicit Euler method by using
C   local and global minimization.
C
C
C   The following subroutines have to be provided to define the 
C   DAE-system:
C
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C     .. Parameters ..
      INTEGER          M, N, NSTEP, LRWORK
      PARAMETER        (M=2, N=3, NSTEP=100, LRWORK=100000)
C     .. Local Scalars ..
      INTEGER          IRED, IERR
      DOUBLE PRECISION T0, TN, H, TI, LERRMAX, GERRMAX
C     .. Local Arrays ..
      INTEGER          CVAL(5), IPAR(1), IOPT(2)
      DOUBLE PRECISION RPAR(1), RWORK(LRWORK), XLREF(N), XLQ(N*NSTEP),
     &                 XGREF(N), XGQ(N*NSTEP), LERR(N), GERR(N), Q(3,3)
C     .. External Subroutines ..
      EXTERNAL         EDEF, ADEF, FDEF     
C     .. External Functions ..
      DOUBLE PRECISION DSECND
      EXTERNAL         DSECND
C     .. Executable Statements ..
      IOUT = 8
      OPEN (IOUT, FILE='main.out')
C   Compute system properties
      T0 = 0.0D0
	TN = 1.0D0
C
C   Do not compute the strangeness index and the reduced system
	IOPT(1) = 1
	IPAR(1) = 1
C   Compute the solution for the standard form system
      CALL DLOCAL(EDEF, ADEF, FDEF, M, N, T0, TN, NSTEP, XLQ, CVAL, 
     $            IPAR, RPAR, RWORK, LRWORK, IOPT, IERR)
      IF (IERR .LT. 0) THEN
	   WRITE (*,*) 'Failed to compute the solution.'
	   STOP
	END IF
	WRITE (*,*) 'Solution by local minimization successfully 
     $computed.'
	CALL DGLOBAL(EDEF, ADEF, FDEF, M, N, T0, TN, NSTEP, XGQ, CVAL,
     &             IPAR, RPAR, RWORK, LRWORK, IOPT, IERR)
	WRITE (*,*) 'Solution by global minimization successfully 
     $computed.'
C
C   Compute the errors and print the results
C
      LERRMAX = 0.0D0
	GERRMAX = 0.0D0
      H = (TN-T0)/NSTEP
      DO 100 ISTEP=1,NSTEP
	   TI = T0 + ISTEP*H
	   XLREF(1) = DEXP(TI) - TI - 1.0D0
	   XLREF(2) = 1.0D0 - 2.0D0*TI - DEXP(TI)
	   XLREF(3) = 0.0D0
	   IF (IPAR(1) .EQ. 1) THEN
	      CALL QDEF(N, TI, Q, N)
	      CALL DGEMV('T', N, N, 1.0D0, Q, N, XLREF, 1, 
     $	          0.0D0, RWORK, 1)
	      CALL DCOPY(N, RWORK, 1, XLREF, 1)
	   END IF
	   LERR(1) = DABS(XLQ((ISTEP-1)*N+1) - XLREF(1))
	   LERR(2) = DABS(XLQ((ISTEP-1)*N+2) - XLREF(2))
	   LERR(3) = DABS(XLQ(ISTEP*N) - XLREF(3))
	   IF (LERR(1) .GT. LERRMAX) LERRMAX = LERR(1)
	   IF (LERR(2) .GT. LERRMAX) LERRMAX = LERR(2)
	   IF (LERR(3) .GT. LERRMAX) LERRMAX = LERR(3)
	   XGREF(1) = DEXP(TI) - 1.0D0
	   XGREF(2) = 1.0D0 - TI - DEXP(TI)
	   XGREF(3) = 1.0D0 - TI
	   IF (IPAR(1) .EQ. 1) THEN
	      CALL QDEF(N, TI, Q, N)
	      CALL DGEMV('T', N, N, 1.0D0, Q, N, XGREF, 1, 
     $	          0.0D0, RWORK, 1)
	      CALL DCOPY(N, RWORK, 1, XGREF, 1)
	   END IF
	   GERR(1) = DABS(XGQ((ISTEP-1)*N+1) - XGREF(1))
	   GERR(2) = DABS(XGQ((ISTEP-1)*N+2) - XGREF(2))
	   GERR(3) = DABS(XGQ(ISTEP*N) - XGREF(3))
	   IF (GERR(1) .GT. GERRMAX) GERRMAX = GERR(1)
	   IF (GERR(2) .GT. GERRMAX) GERRMAX = GERR(2)
	   IF (GERR(3) .GT. GERRMAX) GERRMAX = GERR(3)
         WRITE(IOUT,*)
         WRITE(IOUT,110) TI
         DO 100 I=1,N
            WRITE(IOUT,120) XLREF(I), XLQ((ISTEP-1)*N+I), LERR(I),
     $		              XGREF(I), XGQ((ISTEP-1)*N+I), GERR(I)
 100  CONTINUE  
 110  FORMAT('t=',f10.5,5X,'xl-reference',3X,'xl-computed',4X,
     &       'xl-error',7X,'xg-reference',3X,'xg-computed',4X,
     &       'xg-error')  
 120  FORMAT(15X,3D14.5,3X,3D14.5)
      WRITE (8,130) LERRMAX, GERRMAX
 130  FORMAT(/,/,'Maximum Errors wrt. the infinity norm:',/
     &       5X,'Local Minimization: ',d12.5,/5X,'Global Minimization:'
     &       ,d12.5) 
	WRITE (*,*)
	WRITE (*,*) 'See main.out for results'
      STOP
      END



      SUBROUTINE EDEF(M, N, T, IDIF, ET, LDET, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, N, IDIF, LDET, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION RPAR(*), ET(M,*)
C
C   Evaluate the coefficient function E(t). If IPAR(1)=0, then E has 
C   to correspond with a system in orthogonal standard form.
C   If IPAR(1)=1, then E is multiplied from the right with the 
C   transformation Q defined by the subroutine QDEF.
C
C     .. Local Arrays ..
      DOUBLE PRECISION Q(N,N), E(M,N)
C     .. Local Scalars ..
C     .. External Functions ..
      EXTERNAL         QDEF       
C     .. Executable Statements ..
C
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
C   Enter the coefficients of E(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, ET, M)
      ET(1,1) = 1.0D0
C   Compute transformation Q and replace E by E*Q
      IF (IPAR(1).EQ.1) THEN
         CALL QDEF(N, T, Q, N)
         CALL DGEMM('N','N', M, N, N, 1.0D0, ET, M, Q, N, 0.0D0, E, M)
         CALL DLACPY('A', M, N, E, M, ET, M) 
      END IF
      RETURN
      END

      
      SUBROUTINE ADEF(M, N, T, IDIF, AT, LDAT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, N, IDIF, LDAT, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION RPAR(*), AT(M,*)
C
C   Evaluate the coefficient function A(t). If IPAR(1)=0, then E has 
C   to correspond with a system in orthogonal standard form.
C   If IPAR(1)=1, then E is multiplied from the right with the 
C   transformation Q defined by the subroutine QDEF.
C
C     .. Local Arrays ..
      DOUBLE PRECISION Q(N,N), E(M,N)
C     .. External Functions ..
      EXTERNAL         QDEF, QDOT, EDEF
C     .. Executable Statements ..
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
C   Enter the coefficients of A(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, AT, M)
      AT(1,1) = 1.0D0 - 0.5D0*T
      AT(1,2) = 0.5D0*T
      AT(1,3) = 1.0D0
      AT(2,1) = -1.0D0
      AT(2,2) = 1.0D0
C                                               .
C   Compute transformation Q and its derivative Q
C   and set
C               .
C    A <- A*Q-E*Q.
      IF (IPAR(1).EQ.1) THEN
         CALL QDEF(N, T, Q, N)
         CALL DGEMM('N','N', M, N, N, 1.0D0, AT, M, Q, N, 0.0D0, E, M)
         CALL DLACPY('A', M, N, E, M, AT, M)      
         CALL QDOT(N, T, Q, N)
         CALL EDEF(M, N, T, IDIF, E, M, 0, RPAR, IERR)
         CALL DGEMM('N','N',M, N, N, -1.0D0, E, M, Q, N, 1.0D0, AT, M)
      END IF
      RETURN
      END
       
       
      SUBROUTINE FDEF(M, T, IDIF, FT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, IDIF, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION FT(*), RPAR(*)
C
C   Set the inhomogeneity f(t)
C
C     .. Executable Statements ..
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
      FT(1) = T*(0.5D0*T + DEXP(T))
      FT(2) = T - 2.0D0*(1.0D0 - DEXP(T))
      RETURN
      END
      
     
      SUBROUTINE VDEF(N,T,V)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION V(*)
C 
C   Define the vectors for the Householder transformation Q.
C   Note: You have to initialize exactly N values and V cannot
C         be equal to zero!
C
C     .. Executable Statements ..
      V(1) = 1.0D0*(T+1.0D0)+1.0D0
      V(2) = 1.0D0*T*(T+1.0D0)+1.0D0
      V(3) = 1.0D0
      RETURN
      END
      
      SUBROUTINE VDIF(N,T,VDOT)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION VDOT(*)
C
C   Define the derivatives of the vectors v as defined in the 
C   subroutine VDEF.
C
C     .. Executable Statements ..
      VDOT(1) = 1.0D0
      VDOT(2) = 1.0D0*(2.0D0*T+1.0D0)
      VDOT(3) = 0.0D0
      RETURN
      END
      
      SUBROUTINE QDEF(N, T, Q, LDQ)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Q(LDQ,*)
C
C   Compute the Householder matrix Q=I-2v*v^T/(v^Tv) where v is defined
C   by the subroutine VDEF
C
C     .. Local Scalars ..
      DOUBLE PRECISION VTV
C     .. Local Arrays ..
      DOUBLE PRECISION V(N)
C     .. External Functions ..
      DOUBLE PRECISION DDOT
      EXTERNAL         VDEF, DDOT    
      CALL VDEF(N,T,V)
      CALL DLASET('A',N,N,0.0D0,1.0D0,Q,N)
	VTV = DDOT(N,V,1,V,1)
	IF (VTV .GT. 1.D-12) THEN
         CALL DGEMM('N','T', N, N, 1, -2.0D0/DDOT(N,V,1,V,1), V, N,
     &           V, N, 1.0D0, Q, N)
	END IF
	RETURN
	END

      SUBROUTINE QDOT(N, T, Q, LDQ)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Q(LDQ,*)
C
C   Compute the derivative of the Householder matrix Q=I-2v*v^T/(v^Tv)
C   where v and its derivative are defined by the subroutines VDEF
C   and VDOT.
C
C     .. Local Scalars ..
      DOUBLE PRECISION VTV, VTVDOT
C     .. Local Arrays ..
      DOUBLE PRECISION VVTDOT(N,N), V(N), VPRIME(N)
C     .. External Functions ..
      DOUBLE PRECISION DDOT
      EXTERNAL         VDEF, VDIF, DDOT    
C     .. Executable Statements ..
C                 .
C   Compute v and v
      CALL VDEF(N, T, V)
      CALL VDIF(N, T, VPRIME)
C                                     .
C   Compute v^T*v and d/dt(v^T*v) = 2*v^T*v
      VTV = DDOT(N, V, 1, V, 1)
      VTVDOT = 2.0D0*DDOT(N, VPRIME, 1, V, 1)
C                             .        .
C   Compute Q = d/dt(v*v^T) = v*v^T + (v*v^T)^T
      CALL DGEMM('N','T',N, N, 1, 1.0D0, VPRIME, N, V, N, 0.0D0, Q, LDQ)
      DO 10 I=1,N
         Q(I,I) = Q(I,I) + Q(I,I)
         DO 10 J=1,I-1
	    Q(I,J) = Q(I,J) + Q(J,I)
	    Q(J,I) = Q(I,J)
  10  CONTINUE
C           .
C   Compute Q = (2*v*v^T*d/dt(v^T*v))/(v^T*v)^2 - 2*d/dt(v*v^T)/(v^Tv)
      IF (VTV .GT. 1.D-12) CALL DGEMM('N','T', N, N, 1,
     &           2.0D0*VTVDOT/(VTV*VTV), V, N, V, N,-2.0D0/VTV, Q, N) 
      RETURN
      END
