C=======================================================================
C     Demonstration program for DLMIN
C=======================================================================
C
C   This program solves an over- or underdetermined linear 
C   strangeness-free DAE-system of the form
C         .
C     E(t)x(t) = A(t)x(t) + f(t)
C
C   by solving a discretization with a BDF method where in every step
C   of the discretization the least square solution of the corresponding
C   linear system.
C
C   This algorithm is applied to a system that is given in the orthogonal
C   standard form and to an equivalent system where the solution space
C   is changed by a differentiable transformation.
C
C   A reference solution is computed with the code DGENDA.
C
C
C   The following subroutines have to be provided to define the 
C   DAE-system (see the file setsys.f for an example):
C
C    SETSIZE - User supplied SUBROUTINE
C              This subroutine defines the size of the DAE system and
C              the characteristic values \hat d, \hat a, \hat u and
C              \hat v. It has the form
C                SUBROUTINE SETSIZE(M, N, ID, IA, IU, IV).
C              As output, this subroutine produces the number M of 
C              equations and the number N of variables of the 
C              strangeness-free DAE system. Furthermore it defines 
C              the characteristic values ID=\hat d, IA=\hat a, 
C              IU=\hat u and IV=\hat v of the DAE system.
C
C     SETINT - User supplied SUBROUTINE
C              This subroutine defines the left and right bounds of
C              the integration interval [t_0, T]. It has the form
C                SUBROUTINE SETINT(T0, TN).
C              As output, this subroutine defines the left and the 
C              right interval bounds T0=t_0 and TN=T.
C       
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C              If IPAR(1)=0, then the matrix E has to be defined 
C              according to the orthogonal standard form, i.e. the 
C              leading ID by ID part of E(t) has to be nonsingular,
C              the rest of the matrix has to be equal to zero.
C              If IPAR(1)=1, then the array E contains the matrix
C              E(t)*Q(t) where the orthogonal matrix Q(t) is defined by
C              the subroutine QDEF.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C              If IPAR(1)=0, then the matrix A has to be defined 
C              according to the orthogonal standard form, i.e. 
C              only the blocks A(1:ID,1:N) and A(ID+1:ID+IA,1:ID+IA)
C              contain nonzero entries and the block 
C              A(ID+1,ID+IA,ID+1:ID+IA) is nonsingular.
C              If IPAR(1)=1, then the array A contains the matrix
C              A(t)*Q(t)-E(t)*(d/dt)Q(t) where the orthogonal matrix 
C              Q(t) and its derivative (d/dt)Q(t) are defined by
C              the subroutines QDEF and QDOT.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C     .. Parameters ..
      INTEGER          M, N, NSTEP, LRWORK
      PARAMETER        (M=2, N=2, NSTEP=10, LRWORK=10000, LIWORK=100)
C     .. Local Scalars ..
      INTEGER          IRED, IERR
      DOUBLE PRECISION T0, TN, H, T, LERRMAX, GERRMAX
C     .. Local Arrays ..
      INTEGER          CVAL(5), IPAR(1), IWORK(LIWORK)
      DOUBLE PRECISION RPAR(1), RWORK(LRWORK), XLREF(N), XLQ(N*NSTEP),
     &                 XGREF(N), XGQ(N*NSTEP), LERR(N), GERR(N) 
C     .. External Subroutines ..
      EXTERNAL         EDIF, ADIF, FDIF     
C     .. External Functions ..
      DOUBLE PRECISION DSECND
      EXTERNAL         DSECND
C     .. Executable Statements ..
      IOUT = 8
      OPEN (IOUT, FILE='main_ex1.out')
C   Compute system properties
      T0 = 0.0D0
	TN = 1.0D0
C   Compute the strangeness index and the reduced system
	IRED = 0
C   Compute the solution for the standard form system
      CALL DLOCAL(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XLQ, CVAL, 
     $            IPAR, RPAR, RWORK, LRWORK, IRED, IERR)
      IF (IERR .LT. 0) THEN
	   WRITE (*,*) 'Failed to compute the solution.'
	   STOP
	END IF
	WRITE (*,*) 'Solution by local minimization successfully 
     $computed.'
	CALL DGLOBAL(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XGQ, CVAL,
     &             IPAR, RPAR, RWORK, LRWORK, IRED, IERR)
	WRITE (*,*) 'Solution by global minimization successfully 
     $computed.'
C
C   Compute the errors and print the results
C
      LERRMAX = 0.0D0
	GERRMAX = 0.0D0
      H = (TN-T0)/NSTEP
      DO 100 ISTEP=1,NSTEP
	   T = T0 + ISTEP*H
	   XLREF(1) = (DEXP(T) - T)/(1.0D0 + T*T)
	   XLREF(2) = -T*XLREF(1)
	   LERR(1) = DABS(XLQ((ISTEP-1)*N+1) - XLREF(1))
	   LERR(2) = DABS(XLQ(ISTEP*N) - XLREF(2))
	   IF (LERR(1) .GT. LERRMAX) LERRMAX = LERR(1)
	   IF (LERR(2) .GT. LERRMAX) LERRMAX = LERR(2)
	   IF (LERR(2) .GT. LERRMAX) LERRMAX = LERR(3)
	   XGREF(1) = (DEXP(T) - T)/(1.0D0 + T*T)
	   XGREF(2) = -T*XLREF(1)
	   GERR(1) = DABS(XGQ((ISTEP-1)*N+1) - XGREF(1))
	   GERR(2) = DABS(XGQ(ISTEP*N) - XGREF(2))
	   IF (GERR(1) .GT. GERRMAX) GERRMAX = GERR(1)
	   IF (GERR(2) .GT. GERRMAX) GERRMAX = GERR(2)
         WRITE(IOUT,*)
         WRITE(IOUT,110) T0+ISTEP*H
         DO 100 I=1,N
            WRITE(IOUT,120) XLREF(I), XLQ((ISTEP-1)*N+I), LERR(I),
     $		              XGREF(I), XGQ((ISTEP-1)*N+I), GERR(I)
 100  CONTINUE  
 110  FORMAT('t=',f10.5,5X,'xl-reference',3X,'xl-computed',4X,
     &       'xl-error',7X,'xg-reference',3X,'xg-computed',4X,
     &       'xg-error')  
 120  FORMAT(15X,3D14.5,3X,3D14.5)
      WRITE (8,130) LERRMAX, GERRMAX
 130  FORMAT(/,/,'Maximum Errors wrt. the infinity norm:',/
     &       5X,'Local Minimization: ',d12.5,/5X,'Global Minimization:'
     &       ,d12.5) 
	WRITE (*,*)
	WRITE (*,*) 'Computation successfully finished'
	WRITE (*,*) 'See local.out for results'
      STOP
      END



      SUBROUTINE EDIF(M, N, T, IDIF, ET, LDET, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M,N,LDET
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION ET(LDET,*), RPAR(*)
C
C   Evaluate the coefficient function E(t). 
C
C     .. Executable Statements ..
C
C   Enter the coefficients of E(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, ET, LDET)
	IF (IDIF .EQ. 0) THEN
         ET(1,1) = -T
	   ET(1,2) = T*T
	   ET(2,1) = -1.0D0
	   ET(2,2) = T
	ELSE IF (IDIF .EQ. 1) THEN
         ET(1,1) = -1.0D0
	   ET(1,2) = 2.0D0*T
	   ET(2,1) = 0.0D0
	   ET(2,2) = 1.0D0
      END IF
      RETURN
      END

      
      SUBROUTINE ADIF(M, N, T, IDIF, AT, LDAT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M,N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION AT(LDAT,*), RPAR(*)
C
C   Evaluate the coefficient function A(t).
C
C     .. Executable Statements ..
C   Enter the coefficients of A(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, AT, LDAT)
	IF (IDIF .EQ. 0) THEN
         AT(1,1) = -1.0D0
         AT(1,2) = 0.0D0
         AT(2,1) = 0.0D0
         AT(2,2) = -1.0D0
	END IF
      RETURN
      END
       
       
      SUBROUTINE FDIF(M, T, IDIF, FT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION FT(*)
C
C   Set the inhomogeneity f(t)
C
C     .. Executable Statements ..
	IF (IDIF .EQ. 0) THEN
         FT(1) = DEXP(T)
         FT(2) = 1.0D0
	ELSE IF (IDIF .EQ. 1) THEN
         FT(1) = DEXP(T)
         FT(2) = 0.0D0
	END IF
      RETURN
      END
      
