C=======================================================================
C     Driver routine for DLOCAL, DGLOBAL and ALGO2
C     for the computation of the corresponding execution time
C=======================================================================
C
C   This program solves an over- or underdetermined linear 
C   strangeness-free DAE-system of the form
C         .
C     E(t)x(t) = A(t)x(t) + f(t)
C
C   by solving a discretization with a BDF method where in every step
C   of the discretization the least square solution of the corresponding
C   linear system.
C
C   This algorithm is applied to a system that is given in the orthogonal
C   standard form and to an equivalent system where the solution space
C   is changed by a differentiable transformation.
C
C   A reference solution is computed with the code DGENDA.
C
C
C   The following subroutines have to be provided to define the 
C   DAE-system:
C
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C     .. Parameters ..
      INTEGER          M, N, NSTEP, LRWORK
      PARAMETER        (M=30, N=60, NSTEP=10, LRWORK=10000000)
C     .. Local Scalars ..
      INTEGER          IRED, IERR
      DOUBLE PRECISION T0, TN, DSEC
C     .. Local Arrays ..
      INTEGER          CVAL(5), IPAR(1), INFO(2)
      DOUBLE PRECISION RPAR(1), RWORK(LRWORK), XQ(N*NSTEP)
C     .. External Subroutines ..
      EXTERNAL         EDIF, ADIF, FDIF     
C     .. External Functions ..
      DOUBLE PRECISION DSECND
      EXTERNAL         DSECND
C     .. Executable Statements ..
      IOUT = 8
      OPEN (IOUT, FILE='speedtest.out')
C   Compute system properties
      T0 = 0.0D0
	TN = 1.0D0
C   Do not compute the strangeness index and the reduced system
	IRED = 1
C   Compute the solution using Algorithm 1
      DSEC = DSECND()
      DO 100 I=1,100
         CALL DLOCAL(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XQ, CVAL, 
     $               IPAR, RPAR, RWORK, LRWORK, IRED, IERR)
         IF (IERR .LT. 0) THEN
	      WRITE (*,*) 'Algorithm 1 failed with IERR=', IERR
	      STOP
	   END IF
 	   WRITE (*,*) 'Solution by Algorithm 1 successfully computed.'
 100  CONTINUE
      DSEC = DSECND() - DSEC
	WRITE (IOUT,*) 'Execution time for Algorithm 1:',DSEC/100
	DO 300 I=1,LRWORK
	   RWORK(I) = 0.0D0
 300  CONTINUE
C   Compute the solution using Algorithm 2
      DSEC = DSECND()
	CALL ALGO2(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XQ,
     &           IPAR, RPAR, RWORK, LRWORK, IERR)
      IF (IERR .LT. 0) THEN
	   WRITE (*,*) 'Algorithm 2 failed with IERR=', IERR
	   STOP
	END IF
	WRITE (*,*) 'Solution by Algorithm 2 successfully computed.'
      DSEC = DSECND() - DSEC
	WRITE (IOUT,*) 'Execution time for Algorithm 2:',DSEC
C   Compute the solution using Algorithm 3
      DSEC = DSECND()
	DO 200 I=1,10
	   CALL DGLOBAL(EDIF, ADIF, FDIF, M, N, T0, TN, NSTEP, XQ, CVAL,
     &                IPAR, RPAR, RWORK, LRWORK, IRED, IERR)
         IF (IERR .LT. 0) THEN
	      WRITE (*,*) 'Algorithm 3 failed with IERR=', IERR
	      STOP
	   END IF
	   WRITE (*,*) 'Solution by Algorithm 3 successfully computed.'
 200  CONTINUE
      DSEC = DSECND() - DSEC
	WRITE (IOUT,*) 'Execution time for Algorithm 3:',DSEC/10
      STOP
      END



      SUBROUTINE EDIF(M, N, T, IDIF, ET, LDET, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M,N,LDET
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION ET(LDET,*), RPAR(*)
C
C   Evaluate the coefficient function E(t). 
C
C     .. Executable Statements ..
C
C   Enter the coefficients of E(t)
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, ET, LDET)
      CALL DLASET('A', M, M, 2.0D0, 1.0D0, ET, LDET)
      RETURN
      END

      
      SUBROUTINE ADIF(M, N, T, IDIF, AT, LDAT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M,N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION AT(LDAT,*), RPAR(*)
C
C   Evaluate the coefficient function A(t).
C
C     .. Executable Statements ..
C   Enter the coefficients of A(t)
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
      CALL DLASET('A', M, N, 1.0D0, 2.0D0, AT, LDAT)
      RETURN
      END
       
       
      SUBROUTINE FDIF(M, T, IDIF, FT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION FT(*)
C
C   Set the inhomogeneity f(t)
C
C     .. Executable Statements ..
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
      FT(1) = 1.0D0
      RETURN
      END
      
