C=======================================================================
C     Driver routine for DLOCAL
C=======================================================================
C
C   This program solves an over- or underdetermined linear 
C   strangeness-free DAE-system of the form
C         .
C     E(t)x(t) = A(t)x(t) + f(t)
C
C   by solving a discretization with a BDF method where in every step
C   of the discretization the least square solution of the corresponding
C   linear system.
C
C   This algorithm is applied to a system that is given in the orthogonal
C   standard form and to an equivalent system where the solution space
C   is changed by a differentiable transformation.
C
C   This routine is designed to compare the solutions obtained with BDF
C   methods of different order.
C
C   The following subroutines have to be provided to define the 
C   DAE-system (see the file setsys.f for an example):
C
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C              If IPAR(1)=0, then the matrix E has to be defined 
C              according to the orthogonal standard form, i.e. the 
C              leading ID by ID part of E(t) has to be nonsingular,
C              the rest of the matrix has to be equal to zero.
C              If IPAR(1)=1, then the array E contains the matrix
C              E(t)*Q(t) where the orthogonal matrix Q(t) is defined by
C              the subroutine QDEF.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C              If IPAR(1)=0, then the matrix A has to be defined 
C              according to the orthogonal standard form, i.e. 
C              only the blocks A(1:ID,1:N) and A(ID+1:ID+IA,1:ID+IA)
C              contain nonzero entries and the block 
C              A(ID+1,ID+IA,ID+1:ID+IA) is nonsingular.
C              If IPAR(1)=1, then the array A contains the matrix
C              A(t)*Q(t)-E(t)*(d/dt)Q(t) where the orthogonal matrix 
C              Q(t) and its derivative (d/dt)Q(t) are defined by
C              the subroutines QDEF and QDOT.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C     .. Parameters ..
      INTEGER          M, N, NSTEP, LRWORK
      PARAMETER        (M=2, N=3, NSTEP=100, LRWORK=100)
C     .. Local Scalars ..
      INTEGER          IOUT, IERR
      DOUBLE PRECISION T0, TN, H, TI
C     .. Local Arrays ..
      INTEGER          IPAR(1), IOPT(2)
      DOUBLE PRECISION RWORK(LRWORK), LERR, GERR, XREF(3), XI(3),
     &                 XQ(30000), Q(3,3)       
C     .. External Subroutines ..
      EXTERNAL         EDEF, ADEF, FDEF     
C     .. External Functions ..
      DOUBLE PRECISION DSECND
      EXTERNAL         DSECND
C     .. Executable Statements ..
      IOUT = 8
      OPEN (IOUT, FILE='testbdf.out')
C
C   Setup the interval borders
      T0 = 0.0D0
	TN = 1.0D0
C
C   The rotation is applied to the DAE if IPAR(1) = 1
      IPAR(1) = 1
      IOPT(1) = 1
C
C   Set the desired order of the BDF method
	IORD = 3
	IOPT(2) = IORD
C
C   Setup the initial values
      H = (TN-T0)/NSTEP
	DO 10 I=1,IORD-1
	   TI = T0 + I*H
	   XQ((I-1)*N+1) = DEXP(TI) - TI -1.0D0
	   XQ((I-1)*N+2) = 1.0D0 - 2.0D0*TI - DEXP(TI)
	   XQ((I-1)*N+3) = 0.0D0
	   IF (IPAR(1) .EQ. 1) THEN
	      CALL QDEF(N, TI, Q, N)
	      CALL DGEMV('T', N, N, 1.0D0, Q, N, XQ((I-1)*N+1), 1, 
     $	             0.0D0, XQ(I*N+1), 1)
	      CALL DCOPY(N, XQ(I*N+1), 1, XQ((I-1)*N+1), 1)
	   END IF
  10  CONTINUE
C
C   Compute the solution
      CALL DLOCAL(EDEF, ADEF, FDEF, M, N, T0, TN, NSTEP, XQ, CVAL,
     $           IPAR, RPAR, RWORK, LRWORK, IOPT, IERR)
      IF (IERR .LT. 0) THEN
	   WRITE (*,*) 'Failed to compute the solution.'
	   STOP
	END IF
	WRITE (*,*) 'Local minimization successfully finished.'
C
C   Print the results
      GERR = 0.0D0
	WRITE (IOUT,*) 'local errors'
      DO 100 I=1,NSTEP
	   TI = T0 + I*H
	   XREF(1) = DEXP(TI) - TI -1.0D0
	   XREF(2) = 1.0D0 - 2.0D0*TI - DEXP(TI)
	   XREF(3) = 0.0D0
	   IF (IPAR(1) .EQ. 1) THEN
	      CALL QDEF(N, TI, Q, N)
	      CALL DGEMV('T', N, N, 1.0D0, Q, N, XREF, 1, 
     $	          0.0D0, RWORK, 1)
	      CALL DCOPY(N, RWORK, 1, XREF, 1)
	   END IF
	   CALL DCOPY(N,XQ((I-1)*N+1),1,XI,1)
         LERR = DMAX1(DABS(XI(1)-XREF(1)),DABS(XI(2)-XREF(2)),
     &                   DABS(XI(3)-XREF(3)))
	   WRITE(IOUT,200) TI, LERR
	   IF (LERR .GT. GERR) GERR = LERR
 100  CONTINUE
 200  FORMAT('t=',f6.4,'  lerr=',D12.4)  
	WRITE(IOUT,300) GERR
 300  FORMAT('global error=',D12.4)  
      STOP
      END



      
      SUBROUTINE EDEF(M, N, T, IDIF, ET, LDET, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, N, IDIF, LDET, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION RPAR(*), ET(M,*)
C
C   Evaluate the coefficient function E(t). If IPAR(1)=0, then E has 
C   to correspond with a system in orthogonal standard form.
C   If IPAR(1)=1, then E is multiplied from the right with the 
C   transformation Q defined by the subroutine QDEF.
C
C     .. Local Arrays ..
      DOUBLE PRECISION Q(N,N), E(M,N)
C     .. Local Scalars ..
C     .. External Functions ..
      EXTERNAL         QDEF       
C     .. Executable Statements ..
C
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
C   Enter the coefficients of E(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, ET, M)
      ET(1,1) = 1.0D0
C   Compute transformation Q and replace E by E*Q
      IF (IPAR(1).EQ.1) THEN
         CALL QDEF(N, T, Q, N)
         CALL DGEMM('N','N', M, N, N, 1.0D0, ET, M, Q, N, 0.0D0, E, M)
         CALL DLACPY('A', M, N, E, M, ET, M) 
      END IF
      RETURN
      END

      
      SUBROUTINE ADEF(M, N, T, IDIF, AT, LDAT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, N, IDIF, LDAT, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION RPAR(*), AT(M,*)
C
C   Evaluate the coefficient function A(t). If IPAR(1)=0, then E has 
C   to correspond with a system in orthogonal standard form.
C   If IPAR(1)=1, then E is multiplied from the right with the 
C   transformation Q defined by the subroutine QDEF.
C
C     .. Local Arrays ..
      DOUBLE PRECISION Q(N,N), E(M,N)
C     .. External Functions ..
      EXTERNAL         QDEF, QDOT, EDEF
C     .. Executable Statements ..
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
C   Enter the coefficients of A(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, AT, M)
      AT(1,1) = 1.0D0 - 0.5D0*T
      AT(1,2) = 0.5D0*T
      AT(1,3) = 1.0D0
      AT(2,1) = -1.0D0
      AT(2,2) = 1.0D0
C                                               .
C   Compute transformation Q and its derivative Q
C   and set
C               .
C    A <- A*Q-E*Q.
      IF (IPAR(1).EQ.1) THEN
         CALL QDEF(N, T, Q, N)
         CALL DGEMM('N','N', M, N, N, 1.0D0, AT, M, Q, N, 0.0D0, E, M)
         CALL DLACPY('A', M, N, E, M, AT, M)      
         CALL QDOT(N, T, Q, N)
         CALL EDEF(M, N, T, IDIF, E, M, 0, RPAR, IERR)
         CALL DGEMM('N','N',M, N, N, -1.0D0, E, M, Q, N, 1.0D0, AT, M)
      END IF
      RETURN
      END
       
       
      SUBROUTINE FDEF(M, T, IDIF, FT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, IDIF, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION FT(*), RPAR(*)
C
C   Set the inhomogeneity f(t)
C
C     .. Executable Statements ..
      IF (IDIF .GT. 0) THEN
	   IERR = -2
	   RETURN
	END IF
      FT(1) = T*(0.5D0*T + DEXP(T))
      FT(2) = T - 2.0D0*(1.0D0 - DEXP(T))
      RETURN
      END
      
     
      SUBROUTINE VDEF(N,T,V)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION V(*)
C 
C   Define the vectors for the Householder transformation Q.
C   Note: You have to initialize exactly N values and V cannot
C         be equal to zero!
C
C     .. Executable Statements ..
      V(1) = 1.0D0*(T+1.0D0)+1.0D0
      V(2) = 1.0D0*T*(T+1.0D0)+1.0D0
      V(3) = 1.0D0
      RETURN
      END
      
      SUBROUTINE VDIF(N,T,VDOT)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION VDOT(*)
C
C   Define the derivatives of the vectors v as defined in the 
C   subroutine VDEF.
C
C     .. Executable Statements ..
      VDOT(1) = 1.0D0
      VDOT(2) = 1.0D0*(2.0D0*T+1.0D0)
      VDOT(3) = 0.0D0
      RETURN
      END
      
      SUBROUTINE QDEF(N, T, Q, LDQ)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Q(LDQ,*)
C
C   Compute the Householder matrix Q=I-2v*v^T/(v^Tv) where v is defined
C   by the subroutine VDEF
C
C     .. Local Scalars ..
      DOUBLE PRECISION VTV
C     .. Local Arrays ..
      DOUBLE PRECISION V(N)
C     .. External Functions ..
      DOUBLE PRECISION DDOT
      EXTERNAL         VDEF, DDOT    
      CALL VDEF(N,T,V)
      CALL DLASET('A',N,N,0.0D0,1.0D0,Q,N)
	VTV = DDOT(N,V,1,V,1)
	IF (VTV .GT. 1.D-12) THEN
         CALL DGEMM('N','T', N, N, 1, -2.0D0/DDOT(N,V,1,V,1), V, N,
     &           V, N, 1.0D0, Q, N)
	END IF
	RETURN
	END

      SUBROUTINE QDOT(N, T, Q, LDQ)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Q(LDQ,*)
C
C   Compute the derivative of the Householder matrix Q=I-2v*v^T/(v^Tv)
C   where v and its derivative are defined by the subroutines VDEF
C   and VDOT.
C
C     .. Local Scalars ..
      DOUBLE PRECISION VTV, VTVDOT
C     .. Local Arrays ..
      DOUBLE PRECISION VVTDOT(N,N), V(N), VPRIME(N)
C     .. External Functions ..
      DOUBLE PRECISION DDOT
      EXTERNAL         VDEF, VDIF, DDOT    
C     .. Executable Statements ..
C                 .
C   Compute v and v
      CALL VDEF(N, T, V)
      CALL VDIF(N, T, VPRIME)
C                                     .
C   Compute v^T*v and d/dt(v^T*v) = 2*v^T*v
      VTV = DDOT(N, V, 1, V, 1)
      VTVDOT = 2.0D0*DDOT(N, VPRIME, 1, V, 1)
C                             .        .
C   Compute Q = d/dt(v*v^T) = v*v^T + (v*v^T)^T
      CALL DGEMM('N','T',N, N, 1, 1.0D0, VPRIME, N, V, N, 0.0D0, Q, LDQ)
      DO 10 I=1,N
         Q(I,I) = Q(I,I) + Q(I,I)
         DO 10 J=1,I-1
	    Q(I,J) = Q(I,J) + Q(J,I)
	    Q(J,I) = Q(I,J)
  10  CONTINUE
C           .
C   Compute Q = (2*v*v^T*d/dt(v^T*v))/(v^T*v)^2 - 2*d/dt(v*v^T)/(v^Tv)
      IF (VTV .GT. 1.D-12) CALL DGEMM('N','T', N, N, 1,
     &           2.0D0*VTVDOT/(VTV*VTV), V, N, V, N,-2.0D0/VTV, Q, N) 
      RETURN
      END
