C=======================================================================
C     Demonstration program for LSDAE
C=======================================================================
C
C   This program solves an over- or underdetermined linear 
C   strangeness-free DAE-system of the form
C          .
C     E(t) x(t) = A(t)x(t) + f(t)
C
C   in the least squares sense where the pair (E,A) is computed from a 
C   DAE in the orthogonal standard form by time variant smooth orthogonal 
C   transformations. This solution is computed by solving
C   the minimization problem
C                            
C      ||x_h||_2=min!  s.t. ||D_h*x_h-f_h||_2=min!
C
C   where the linear system
C
C      D_h*x_h = f_h
C
C   represents the discretization of the above differential algebraic
C   system with the implicit Euler method and with a homogegeous 
C   initial value. This system is solved with respect to the 
C   block bidiagonal structure of the matrix D_h.
C
C   A reference solution is computed as the solution of the boundary value
C   problem associated with the corresponding control problem.
C
C   To define the problem the following subroutines have to be provided
C   by the user (see the file setsys.f for an example):
C
C    SETSIZE - User supplied SUBROUTINE
C              This subroutine defines the size of the DAE system and
C              the characteristic values \hat d, \hat a, \hat u and
C              \hat v. It has the form
C                SUBROUTINE SETSIZE(M, N, ID, IA, IU, IV).
C              As output, this subroutine produces the number M of 
C              equations and the number N of variables of the 
C              strangeness-free DAE system. Furthermore it defines 
C              the characteristic values ID=\hat d, IA=\hat a, 
C              IU=\hat u and IV=\hat v of the DAE system.
C
C     SETINT - User supplied SUBROUTINE
C              This subroutine defines the left and right bounds of
C              the integration interval [t_0, T]. It has the form
C                SUBROUTINE SETINT(T0, TN).
C              As output, this subroutine defines the left and the 
C              right interval bounds T0=t_0 and TN=T.
C       
C       EDEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix E(t). 
C              It is of the form
C                SUBROUTINE EDEF(M, N, T, E, LDE, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array E of dimension (LDE,N) where the leading M by N
C              part contains the matrix E(t) at the time T.
C              If IPAR(1)=0, then the matrix E has to be defined 
C              according to the orthogonal standard form, i.e. the 
C              leading ID by ID part of E(t) has to be nonsingular,
C              the rest of the matrix has to be equal to zero.
C              If IPAR(1)=1, then the array E contains the matrix
C              E(t)*Q(t) where the orthogonal matrix Q(t) is defined by
C              the subroutine QDEF.
C
C       ADEF - User supplied SUBROUTINE
C              This subroutine defines the coefficient matrix A(t). 
C              It is of the form
C                SUBROUTINE ADEF(M, N, T, A, LDA, IPAR).
C              The subroutine takes as input the number M of equations
C              and the number N of variables of the DAE system. As 
C              output the subroutine produces the DOUBLE PRECISION
C              array A of dimension (LDE,N) where the leading M by N
C              part contains the matrix A(t) at the time T.
C              If IPAR(1)=0, then the matrix A has to be defined 
C              according to the orthogonal standard form, i.e. 
C              only the blocks A(1:ID,1:N) and A(ID+1:ID+IA,1:ID+IA)
C              contain nonzero entries and the block 
C              A(ID+1,ID+IA,ID+1:ID+IA) is nonsingular.
C              If IPAR(1)=1, then the array A contains the matrix
C              A(t)*Q(t)-E(t)*(d/dt)Q(t) where the orthogonal matrix 
C              Q(t) and its derivative (d/dt)Q(t) are defined by
C              the subroutines QDEF and QDOT.
C
C       FDEF - User supplied SUBROUTINE
C              This subroutine defines the inhomogeneity f(t). 
C              It is of the form
C                SUBROUTINE FDEF(M, T, F).
C              The subroutine takes as input the number M of equations
C              and the time T. As output the subroutine produces the 
C              DOUBLE PRECISION vector F. The first M entries of this
C              vector contain the inhomogeneity f(t).
C
C
C     .. Parameters ..
      INTEGER          M, N, NEQNS, NSTEP, LRWORK
      PARAMETER        (NSTEP=100, LRWORK=300000)
C     .. Local Scalars ..
      INTEGER          IERR
      DOUBLE PRECISION T0, TN, H, XDIFF1, DIFF1, XDIFF2, DIFF2, 
     &                 XDIFF3, DIFF3, DSEC
C     .. Local Arrays ..
      INTEGER          IPAR(1), IOPT(1), CVAL(5)
      DOUBLE PRECISION RWORK(LRWORK), XQREF(30000),
     &                 XQ(30000), XQNF(30000), RPAR(1)       
C     .. External Functions ..
      DOUBLE PRECISION DSECND
      EXTERNAL         DSECND
C     .. External Subroutines ..
      EXTERNAL         EDEF, ADEF, FDEF
C     .. Executable Statements ..
      IOUT = 8
      OPEN (IOUT, FILE='global.out')
C
C   Compute system properties
      CALL SETSIZE(M, N, ID, IA, IU, IV)
      CALL SETINT(T0, TN)
      IF (IU .NE. 0) THEN
         NEQNS = 2*ID
      ELSE
         NEQNS = ID
      END IF
C
C   Compute the reference solution
      CALL SLVBVP(NEQNS, NSTEP, XQREF, RWORK, LRWORK)
	WRITE (*,*) 'Reference solution successfully computed.'
C
C   Set the IOPT-parameters.
C   IOPT(1)=1 --> do not compute the strangeness index.
      IOPT(1) = 1
C
C   Compute the solution for the rotated system.
      IPAR(1) = 1
      DSEC = DSECND()
      CALL DGLOBAL(EDEF, ADEF, FDEF, M, N, T0, TN, NSTEP, XQ, CVAL,
     $             IPAR, RPAR, RWORK, LRWORK, IOPT, IERR)
      IF (IERR .LT. 0) THEN
	   WRITE (*,*) 'Failed to compute the solution.'
	   STOP
	END IF
      DSEC = DSECND()-DSEC
	WRITE (*,*) 'Rotated solution successfully computed.'
C
C   Apply rotation to the result.
      CALL XROT(N, T0, TN, NSTEP, XQ)
C
C   Compute the solution for the standard form system.
      IPAR(1) = 0
      CALL DGLOBAL(EDEF, ADEF, FDEF, M, N, T0, TN, NSTEP, XQNF, CVAL,
     $             IPAR, RPAR, RWORK, LRWORK, IOPT,  IERR)
      IF (IERR .LT. 0) THEN
	   WRITE (*,*) 'Failed to compute the solution.'
	   STOP
	END IF
	WRITE (*,*) 'Standard form solution successfully computed.'
C
C   Compute the differences of the solutions
      XDIFF1 = 0.0D0
      XDIFF2 = 0.0D0
      XDIFF3 = 0.0D0
      DO 10 I=1,NSTEP*N
         DIFF1 = DABS(XQ(I)-XQREF(I))
	 DIFF2 = DABS(XQNF(I)-XQREF(I))
	 DIFF3 = DABS(XQ(I)-XQNF(I))
	 IF (DIFF1 .GT. XDIFF1) XDIFF1 = DIFF1
	 IF (DIFF2 .GT. XDIFF2) XDIFF2 = DIFF2
	 IF (DIFF3 .GT. XDIFF3) XDIFF3 = DIFF3
  10  CONTINUE
C
C   Print the results
      H = (TN-T0)/NSTEP
      DO 100 I=1,NSTEP
         WRITE(IOUT,*)
         WRITE(IOUT,110) T0+I*H
         DO 100 J=1,N
            WRITE(IOUT,120) XQREF((I-1)*N+J),XQ((I-1)*N+J),
     &                      XQNF((I-1)*N+J), 
     &                      DABS(XQREF((I-1)*N+J)-XQ((I-1)*N+J)),
     &                      DABS(XQREF((I-1)*N+J)-XQNF((I-1)*N+J)),
     &                      DABS(XQNF((I-1)*N+J)-XQ((I-1)*N+J))
 100  CONTINUE  
 110  FORMAT('t=',f10.5,5X,'x-reference',4X,'x-rotated',5X,'x-normform',
     &       7X,'|xref-xrot|',2X,'|xref-xnform|',1X,'|xnform-xrot|')  
 120  FORMAT(15X,3D14.5,3X,3D14.5)
      WRITE (8,130) XDIFF1, XDIFF2, XDIFF3, DSEC
 130  FORMAT(/,/,'Maximum Errors wrt. the infinity norm:',/
     &       5X,'||xref - xrot||=',f12.7,/5X,'||xref-xnform||=',f12.7
     &       ,/5X,'||xnform-xrot||=',f12.7//
     &       'Computation time for the MP-solver= ',f10.5 ) 
	WRITE (*,*)
	WRITE (*,*) 'See global.out for results'
      STOP
      END


      SUBROUTINE SETSIZE(M, N, ID, IA, IU, IV)
C     .. Scalar Arguments .. 
      INTEGER M, N, ID, IA, IU, IV
C
C   Set system size and characteristic values
C    M   Number of rows of the system / Number of equations
C    N   Number of columns of the system / Number of variables
C    ID  Number of differential equations
C    IA  Number of algebraic equations
C    IU  Number of undetermined solution components
C    IV  Number of vanishing equations
C
C     .. Executable Statements ..
      M=2
      N=3
      ID=1
      IA=1
      IU=1
      IV=0
      IF (ID+IA+IU .NE. N .OR. ID+IA+IV .NE. M) THEN
         WRITE(*,*) 'Theres something wrong with the characteristic',
     $              ' values' 
         STOP
      END IF
      RETURN
      END
      
      
      SUBROUTINE SETINT(A, B)
C     .. Scalar Arguments .. 
      DOUBLE PRECISION A, B
C
C   Set the endpoints of the problem interval
C
C     .. Executable Statements ..
      A = 0.0D0
      B = 1.0D0
      RETURN
      END
      
      
      SUBROUTINE EDEF(M, N, T, IDIF, ET, LDET, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, N, IDIF, LDET, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION ET(LDET,*), RPAR(*)
C
C   Evaluate the coefficient function E(t). If IPAR(1)=0, then E has 
C   to correspond with a system in orthogonal standard form.
C   If IPAR(1)=1, then E is multiplied from the right with the 
C   transformation Q defined by the subroutine QDEF.
C
C     .. Local Arrays ..
      DOUBLE PRECISION Q(N,N), E(M,N)
C     .. Local Scalars ..
C     .. External Functions ..
      EXTERNAL         QDEF       
C     .. Executable Statements ..
C
C   Enter the coefficients of E(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, ET, LDET)
C   ET(1:ID,1:ID) = Sigma_E(t)
C   Note: Sigma_E(t) has to be nonsingular.
      ET(1,1) = 1.0D0
C   Compute transformation Q and replace E by E*Q
      IF (IPAR(1).EQ.1) THEN
         CALL QDEF(N, T, Q, N)
         CALL DGEMM('N','N', M, N, N, 1.0D0, ET, LDET, Q, N, 0.0D0, 
     $              E, M)
         CALL DLACPY('A', M, N, E, M, ET, LDET) 
      END IF
      RETURN
      END

      
      SUBROUTINE ADEF(M, N, T, IDIF, AT, LDAT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, N, IDIF, LDAT
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION AT(LDAT,*), RPAR(*)
C
C   Evaluate the coefficient function A(t). If IPAR(1)=0, then E has 
C   to correspond with a system in orthogonal standard form.
C   If IPAR(1)=1, then E is multiplied from the right with the 
C   transformation Q defined by the subroutine QDEF.
C
C     .. Local Scalars ..
      DOUBLE PRECISION RDUMMY
C     .. Local Arrays ..
      DOUBLE PRECISION Q(N,N), E(M,N)
C     .. External Functions ..
      EXTERNAL         QDEF, QDOT, EDEF
C     .. Executable Statements ..
C   Enter the coefficients of A(t)
      CALL DLASET('A', M, N, 0.0D0, 0.0D0, AT, LDAT)
C   AT(1:ID,1:ID) = A_{11}(t)
      AT(1,1) = 1.0D0 - 0.5D0*T
C   AT(1:ID,ID+1:ID+IA) = A_{12}(t)
      AT(1,2) = 0.5D0*T
C   AT(1:ID,ID+IA+1:ID+IA+IU) = A_{13}(t)
      AT(1,3) = 1.0D0
C   AT(ID+1:ID+IA,1:ID) = A_{21}(t)
      AT(2,1) = -1.0D0
C   AT(ID+1:ID+IA,ID+1:ID+IA) = Sigma_A(t)
C   Note: Sigma_A(t) has to be nonsingular.
      AT(2,2) = 1.0D0
C                                               .
C   Compute transformation Q and its derivative Q
C   and set           .
C    A <- A*Q-E*Q.
      IF (IPAR(1).EQ.1) THEN
         CALL QDEF(N, T, Q, N)
         CALL DGEMM('N','N', M, N, N, 1.0D0, AT, LDAT, Q, N, 0.0D0, 
     $              E, M)
         CALL DLACPY('A', M, N, E, M, AT, LDAT)      
         CALL QDOT(N, T, Q, N)
         CALL EDEF(M, N, T, 0, E, M, 0, RDUMMY, IERR)
         CALL DGEMM('N','N',M, N, N, -1.0D0, E, M, Q, N, 1.0D0, 
     $              AT, LDAT)
      END IF
      RETURN
      END
       
       
      SUBROUTINE FDEF(M, T, IDIF, FT, IPAR, RPAR, IERR)
C     .. Scalar Arguments .. 
      INTEGER          M, IDIF, IERR
      DOUBLE PRECISION T
C     .. Array Arguments ..
      INTEGER          IPAR(*)
      DOUBLE PRECISION FT(*), RPAR(*)
C
C   Set the inhomogeneity f(t)
C
C     .. Executable Statements ..
C   FT(1:ID) = f_1(t)
      FT(1) = T*(0.5D0*T + DEXP(T))
C   FT(ID+1:ID+IA) = f_2(t)
      FT(2) = T - 2.0D0*(1.0D0 - DEXP(T))
C   FT(ID+IA+1:ID+IA+IV) = f_3(t)
      RETURN
      END
      
     
      SUBROUTINE VDEF(N,T,V)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION V(*)
C 
C   Define the vectors for the Householder transformation Q.
C   Note: You have to initialize exactly N values and V cannot
C         be equal to zero!
C
C     .. Executable Statements ..
      V(1) = 1.0D0*(T+1.0D0)+1.0D0
      V(2) = 1.0D0*T*(T+1.0D0)+1.0D0
      V(3) = 1.0D0
      RETURN
      END
      
      SUBROUTINE VDIF(N,T,VDOT)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION VDOT(*)
C
C   Define the derivatives of the vectors v as defined in the 
C   subroutine VDEF.
C
C     .. Executable Statements ..
      VDOT(1) = 1.0D0
      VDOT(2) = 1.0D0*(2.0D0*T+1.0D0)
      VDOT(3) = 0.0D0
      RETURN
      END
      
      SUBROUTINE QDEF(N, T, Q, LDQ)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Q(LDQ,*)
C
C   Compute the Householder matrix Q=I-2v*v^T/(v^Tv) where v is defined
C   by the subroutine VDEF
C
C     .. Local Scalars ..
      DOUBLE PRECISION VTV
C     .. Local Arrays ..
      DOUBLE PRECISION V(N)
C     .. External Functions ..
      DOUBLE PRECISION DDOT
      EXTERNAL         VDEF, DDOT    
      CALL VDEF(N,T,V)
      CALL DLASET('A',N,N,0.0D0,1.0D0,Q,N)
	VTV = DDOT(N,V,1,V,1)
	IF (VTV .GT. 1.D-12) THEN
         CALL DGEMM('N','T', N, N, 1, -2.0D0/DDOT(N,V,1,V,1), V, N,
     &           V, N, 1.0D0, Q, N)
	END IF
	RETURN
	END

      SUBROUTINE QDOT(N, T, Q, LDQ)
C     .. Scalar Arguments .. 
      INTEGER          N
      DOUBLE PRECISION T
C     .. Array Arguments ..
      DOUBLE PRECISION Q(LDQ,*)
C
C   Compute the derivative of the Householder matrix Q=I-2v*v^T/(v^Tv)
C   where v and its derivative are defined by the subroutines VDEF
C   and VDOT.
C
C     .. Local Scalars ..
      DOUBLE PRECISION VTV, VTVDOT
C     .. Local Arrays ..
      DOUBLE PRECISION VVTDOT(N,N), V(N), VPRIME(N)
C     .. External Functions ..
      DOUBLE PRECISION DDOT
      EXTERNAL         VDEF, VDIF, DDOT    
C     .. Executable Statements ..
C                 .
C   Compute v and v
      CALL VDEF(N, T, V)
      CALL VDIF(N, T, VPRIME)
C                                     .
C   Compute v^T*v and d/dt(v^T*v) = 2*v^T*v
      VTV = DDOT(N, V, 1, V, 1)
      VTVDOT = 2.0D0*DDOT(N, VPRIME, 1, V, 1)
C                             .        .
C   Compute Q = d/dt(v*v^T) = v*v^T + (v*v^T)^T
      CALL DGEMM('N','T',N, N, 1, 1.0D0, VPRIME, N, V, N, 0.0D0, Q, LDQ)
      DO 10 I=1,N
         Q(I,I) = Q(I,I) + Q(I,I)
         DO 10 J=1,I-1
	    Q(I,J) = Q(I,J) + Q(J,I)
	    Q(J,I) = Q(I,J)
  10  CONTINUE
C           .
C   Compute Q = (2*v*v^T*d/dt(v^T*v))/(v^T*v)^2 - 2*d/dt(v*v^T)/(v^Tv)
      IF (VTV .GT. 1.D-12) CALL DGEMM('N','T', N, N, 1,
     &           2.0D0*VTVDOT/(VTV*VTV), V, N, V, N,-2.0D0/VTV, Q, N) 
      RETURN
      END
